import h5py
import nibabel
import numpy as np
from tensorflow.keras.utils import to_categorical

from ml4h.normalizer import ZeroMeanStd1
from ml4h.metrics import weighted_crossentropy
from ml4h.TensorMap import TensorMap, Interpretation
from ml4h.tensormap.general import get_tensor_at_first_date, normalized_first_date, pad_or_crop_array_to_shape


t1_slice_80 = TensorMap(
    'axial_80',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)
t1_mni_slice_48 = TensorMap(
    'axial_48',
    shape=(176, 216, 1),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)

t1_mni_slice_80 = TensorMap(
    'axial_80',
    shape=(176, 216, 1),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)
t1_slice_85 = TensorMap(
    'axial_85',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)

t1_slice_100 = TensorMap(
    'axial_100',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)
t1_slice_100_224 = TensorMap(
    'axial_100',
    shape=(224, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)
t1_slice_120 = TensorMap(
    'axial_120',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)
t1_slice_160 = TensorMap(
    'axial_160',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=normalized_first_date,
    normalization=ZeroMeanStd1(),
)


def _brain_volume_from_file(tm, hd5, dependents={}):
    tensor = np.zeros(tm.shape, dtype=np.float32)
    begin_slice = int(tm.name.split('_')[-2])
    end_slice = int(tm.name.split('_')[-1])
    for i in range(begin_slice, end_slice):
        slicer = get_tensor_at_first_date(hd5, tm.path_prefix, f'axial_{i}')
        tensor[..., i-begin_slice] = pad_or_crop_array_to_shape((tm.shape[0], tm.shape[1]), slicer)
    return tensor


t1_slices_68_100 = TensorMap(
    'axial_68_100',
    shape=(216, 256, 32),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)

t1_slices_68_100 = TensorMap(
    'axial_68_100',
    shape=(216, 256, 32),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)

t1_mni_slices_16_48 = TensorMap(
    'axial_16_48',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_48_80 = TensorMap(
    'axial_48_80',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_80_144 = TensorMap(
    'axial_80_144',
    shape=(176, 216, 64),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_0_182 = TensorMap(
    'axial_0_182',
    shape=(176, 216, 182),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_0_32 = TensorMap(
    'axial_0_32',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)

t1_mni_slices_32_64 = TensorMap(
    'axial_32_64',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_64_96 = TensorMap(
    'axial_64_96',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_96_128 = TensorMap(
    'axial_96_128',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_128_160 = TensorMap(
    'axial_128_160',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_30_60 = TensorMap(
    'axial_30_60',
    shape=(176, 216, 30),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)
t1_mni_slices_10_73 = TensorMap(
    'axial_10_73',
    shape=(176, 216, 63),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_brain_volume_from_file,
    normalization=ZeroMeanStd1(),
)

def _segmented_brain_tensor_from_file(tm, hd5, dependents={}):
    # from mapping given in https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FIRST/UserGuide
    num2idx = {0: 0, 10: 1, 11: 2, 12: 3, 13: 4, 16: 5, 17: 6, 18: 7, 26: 8, 49: 9, 50: 10, 51: 11, 52: 12, 53: 13, 54: 14, 58: 15}
    tensor = np.zeros(tm.shape, dtype=np.float32)
    if tm.axes() == 3:
        categorical_index_slice = get_tensor_at_first_date(hd5, tm.path_prefix, tm.name)
        index_remap = np.zeros_like(categorical_index_slice)
        for key in num2idx:
            index_remap[categorical_index_slice == key] = num2idx[key]
        categorical_one_hot = to_categorical(index_remap, len(tm.channel_map))
        tensor[..., :] = pad_or_crop_array_to_shape(tensor[..., :].shape, categorical_one_hot)
    else:
        raise ValueError(f'No method to get segmented slices for TensorMap: {tm}')
    return tensor


brain_channel_map = {
    'Background': 0, 'Left_Thalamus_Proper': 1, 'Left_Caudate': 2, 'Left_Putamen': 3, 'Left_Pallidum': 4, 'Brain_Stem': 5, 'Left_Hippocampus': 6,
    'Left_Amygdala': 7, 'Left_Accumbens_area': 8, 'Right_Thalamus_Proper': 9, 'Right_Caudate': 10, 'Right_Putamen': 11, 'Right_Pallidum': 12,
    'Right_Hippocampus': 13, 'Right_Amygdala': 14, 'Right_Accumbens_area': 15,
}

t1_seg_slice_80 = TensorMap(
    'axial_80',
    interpretation=Interpretation.CATEGORICAL,
    shape=(216, 256, len(brain_channel_map)),
    path_prefix='ukb_brain_mri/T1_first_all_fast_firstseg/',
    channel_map=brain_channel_map,
    loss=weighted_crossentropy([0.1]+[10.0]*15),
    tensor_from_file=_segmented_brain_tensor_from_file,
)

t1_seg_slice_85 = TensorMap(
    'axial_85',
    interpretation=Interpretation.CATEGORICAL,
    shape=(216, 256, len(brain_channel_map)),
    path_prefix='ukb_brain_mri/T1_first_all_fast_firstseg/',
    channel_map=brain_channel_map,
    tensor_from_file=_segmented_brain_tensor_from_file,
)


def _brain_label_masked(labels, segmentation_key='ukb_brain_mri/T1_first_all_fast_firstseg/'):
    def _masked_brain_tensor(tm, hd5, dependents={}):
        # from mapping given in https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FIRST/UserGuide
        num2idx = {0: 0, 10: 1, 11: 2, 12: 3, 13: 4, 16: 5, 17: 6, 18: 7, 26: 8, 49: 9, 50: 10, 51: 11, 52: 12, 53: 13, 54: 14, 58: 15}
        tensor = np.zeros(tm.shape, dtype=np.float32)
        if tm.axes() == 3:
            mri = get_tensor_at_first_date(hd5, tm.path_prefix, tm.name)
            categorical_index_slice = get_tensor_at_first_date(hd5, segmentation_key, tm.name)
            index_remap = np.zeros_like(categorical_index_slice)
            for key in num2idx:
                index_remap[categorical_index_slice == key] = num2idx[key]

            label_mask = np.isin(index_remap, list(labels.values()))
            mri = label_mask * mri
            tensor = pad_or_crop_array_to_shape(tensor.shape, mri)
        else:
            raise ValueError(f'No method to get segmented slices for TensorMap: {tm}')
        return tensor
    return _masked_brain_tensor


t1_slice_80_hippocampus = TensorMap(
    'axial_80',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=_brain_label_masked({'Left_Hippocampus': 6, 'Right_Hippocampus': 13}),
    normalization=ZeroMeanStd1(),
)
t1_slice_92_putamen = TensorMap(
    'axial_92',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=_brain_label_masked({'Left_Putamen': 3, 'Right_Putamen': 11}),
    normalization=ZeroMeanStd1(),
)
t1_slice_63_brainstem = TensorMap(
    'axial_63',
    shape=(216, 256, 1),
    path_prefix='ukb_brain_mri/T1/',
    tensor_from_file=_brain_label_masked({'Brain_Stem': 5}),
    normalization=ZeroMeanStd1(),
)


def _mni_label_masked(labels, mni_label_mask='/home/sam/mni_icbm152_CerebrA_tal_nlin_sym_09c.nii'):
    mni_nifti = nibabel.load('/home/sam/mni_icbm152_CerebrA_tal_nlin_sym_09c.nii')
    mni_labels = mni_nifti.get_fdata()[:176, :216, :182]  # crop to UKB MNI

    def _masked_brain_tensor(tm, hd5, dependents={}):
        tensor = np.zeros(tm.shape, dtype=np.float32)
        begin_slice = int(tm.name.split('_')[-2])
        end_slice = int(tm.name.split('_')[-1])
        for i in range(begin_slice, end_slice):
            slicer = get_tensor_at_first_date(hd5, tm.path_prefix, f'axial_{i}')
            tensor[..., i - begin_slice] = pad_or_crop_array_to_shape((tm.shape[0], tm.shape[1]), slicer)
        label_mask = np.isin(mni_labels[..., begin_slice:end_slice], list(labels.values()))
        tensor *= label_mask
        return tensor
    return _masked_brain_tensor


t1_mni_slice_80_hippocampus = TensorMap(
    'mni_hippocampus_axial_48_80',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_Hippocampus': 99, 'Right_Hippocampus': 48}),  # CerebrA Label Map
    normalization=ZeroMeanStd1(),
)
t1_mni_slice_60_92_putamen = TensorMap(
    'mni_putamen_axial_60_92',
    shape=(176, 216, 32),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_Putamen': 72, 'Right_Putamen': 21}),  # CerebrA Label Map
    normalization=ZeroMeanStd1(),
)
t1_mni_cerebellum_white_matter_30_60 = TensorMap(
    't1_mni_cerebellum_white_matter_30_60',
    shape=(176, 216, 30),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_cerebellum_white_matter': 90, 'Right_cerebellum_white_matter': 39}),
    normalization=ZeroMeanStd1(),
)
t1_mni_cerebellum_gray_matter_30_60 = TensorMap(
    't1_mni_cerebellum_gray_matter_30_60',
    shape=(176, 216, 30),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_cerebellum_gray_matter': 97, 'Right_cerebellum_gray_matter': 46}),
    normalization=ZeroMeanStd1(),
)
t1_mni_cerebellum_30_60 = TensorMap(
    't1_mni_cerebellum_30_60',
    shape=(176, 216, 30),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({
        'Left_cerebellum_gray_matter': 97, 'Right_cerebellum_gray_matter': 46,
        'Left_cerebellum_white_matter': 90, 'Right_cerebellum_white_matter': 39,
    }),
    normalization=ZeroMeanStd1(),
)
t1_mni_cerebellum_gray_matter_10_73 = TensorMap(
    't1_mni_cerebellum_gray_matter_10_73',
    shape=(176, 216, 63),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_cerebellum_gray_matter': 97, 'Right_cerebellum_gray_matter': 46}),
    normalization=ZeroMeanStd1(),
)
t1_mni_cerebellum_gray_matter_10_73 = TensorMap(
    't1_mni_cerebellum_gray_matter_10_73',
    shape=(176, 216, 63),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_cerebellum_gray_matter': 97, 'Right_cerebellum_gray_matter': 46}),
    normalization=ZeroMeanStd1(),
)
t1_mni_cerebellum_10_73 = TensorMap(
    't1_mni_cerebellum_10_73',
    shape=(176, 216, 63),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({
        'Left_cerebellum_gray_matter': 97, 'Right_cerebellum_gray_matter': 46,
        'Left_cerebellum_white_matter': 90, 'Right_cerebellum_white_matter': 39,
    }),
    normalization=ZeroMeanStd1(),
)
t1_mni_amygdala_47_71 = TensorMap(
    't1_mni_amygdala_47_71',
    shape=(176, 216, 24),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_amygdala': 19, 'Right_amygdala': 70}),
    normalization=ZeroMeanStd1(),
)
t1_mni_accumbens_65_76 = TensorMap(
    't1_mni_accumbens_65_76',
    shape=(176, 216, 11),
    path_prefix='ukb_brain_mri/T1_brain_to_MNI/',
    tensor_from_file=_mni_label_masked({'Left_accumbens': 19, 'Right_accumbens': 70}),
    normalization=ZeroMeanStd1(),
)



from ml4ht.data.defines import SampleID

import os
import glob
from typing import Callable, List, Union, Optional, Tuple, Dict, Any

import h5py
import numcodecs
import numpy as np
import pandas as pd

from ml4ht.data.data_description import DataDescription
from ml4ht.data.util.date_selector import DATE_OPTION_KEY
from ml4ht.data.defines import LoadingOption, Tensor

from ml4h.TensorMap import TensorMap
from ml4h.defines import PARTNERS_DATETIME_FORMAT, ECG_REST_AMP_LEADS


def sample_id_from_path(hd5_path: str) -> int:
    return int(os.path.splitext(os.path.basename(hd5_path))[0])


# this is used to load data from our hd5 storage
def decompress_data(data_compressed: np.array, dtype: str) -> np.array:
    """Decompresses a compressed byte array. If the primitive type of the data
    to decompress is a string, calls decode using the zstd codec. If the
    primitive type of the data to decompress is not a string (e.g. int or
    float), the buffer is interpreted using the passed dtype."""
    codec = numcodecs.zstd.Zstd()
    data_decompressed = codec.decode(data_compressed)
    if dtype == "str":
        data = data_decompressed.decode()
    else:
        data = np.frombuffer(data_decompressed, dtype)
    return data


class ECGDataDescription(DataDescription):
    S3_PATH_OPTION = "s3_path"
    TEXT_DIAGNOSES = {
        "sb": ["sinus brady"],
        "st": ["sinus tachy"],
        "af": [
            "afib",
            "atrial fib",
            "afibrillation",
            "atrialfibrillation",
        ],
        "rbbb": [
            "right bbb",
            "rbbb",
            "right bundle branch block",
        ],
        "lbbb": [
            "left bbb",
            "lbbb",
            "left bundle branch block",
        ],
        "avb": [
            "1st degree atrioventricular block",
            "1st degree av block",
            "first degree av block",
            "first degree atrioventricular block",
        ],
        "lvh": [
            "biventricular hypertrophy",
            "biventriclar hypertrophy",
            "leftventricular hypertrophy",
            "combined ventricular hypertrophy",
            "left ventricular hypertr",
        ],
        "ischemia": [
            "diffuse st segment elevation",
            "consistent with lateral ischemia",
            "subendocardial ischemia",
            "apical subendocardial ischemia",
            "inferior subendocardial ischemia",
            "anterolateral ischemia",
            "antero-apical ischemia",
            "consider anterior and lateral ischemia",
            "st segment depression",
            "minor st segment depression",
            "st segment depression in leads v4-v6",
            "anterolateral st segment depression",
            "infero- st segment depression",
            "st depression",
            "suggest anterior ischemia",
            "st segment depression is more marked in leads",
            "possible anterior wall ischemia",
            "consistent with ischemia",
            "diffuse scooped st segment depression",
            "anterolateral subendocardial ischemia",
            "diffuse st segment depression",
            "st segment elevation consistent with acute injury",
            "inferior st segment elevation and q waves",
            "st segment depression in anterolateral leads",
            "widespread st segment depression",
            "consider anterior ischemia",
            "suggesting anterior ischemia",
            "consistent with subendocardial ischemia",
            "marked st segment depression in leads",
            "inferior st segment depression",
            "st segment elevation in leads",
            "st segment elevation",
            "st segment depressions more marked",
            "anterior st segment depression",
            "apical st depression",
            "septal ischemia",
            "st segment depression in leads",
            "suggests anterolateral ischemia",
            "st elevation",
            "diffuse elevation of st segments",
            "marked st segment depression",
            "anterior infarct or transmural ischemia",
            "inferoapical st segment depression",
            "lateral ischemia",
            "nonspecific st segment depression",
            "anterior subendocardial ischemia",
        ],
    }
    META_DATA_FIELDS = [
        "gender",
        "patientage",
        "locationname",
        "waveform_samplebase",
        "ventricularrate_md",
        "qrsduration_md",
        "printerval_md",
        "qtinterval_md",
        "paxis_md",
        "raxis_md",
        "taxis_md",
        "weightlbs",
        "heightin",
    ]

    def __init__(
            self,
            local_hd5_folder: str,
            name: str,
            ecg_len: int,
            transforms: List[Callable[[Tensor, LoadingOption], Tensor]] = None,
            date_format: str = PARTNERS_DATETIME_FORMAT,
            hd5_path_to_ecg: str = "partners_ecg_rest",
            leads: Dict[str, int] = ECG_REST_AMP_LEADS,
    ):
        """
        Reads ECGs in our hd5 format stored on ccds s3.
        If s3 information is provided, if an hd5 is not available locally,
        it's downloaded from s3.
        :param local_hd5_folder: Where to check for locally stored hd5s.
            Also where hd5s will be downloaded to from s3.
        :param name: Name of the output of this DataDescription.
        :param ecg_len: length in samples to interpolate all leads of ECG to.
        :param transforms: transformations including augmentations to apply to ECG.
        :param hd5_path_to_ecg: key in hd5 of the ECG leads
        :param leads: mapping from lead name in hd5 -> channel in output array
        """
        self.local_hd5_folder = local_hd5_folder
        self.date_format = date_format
        self.hd5_path_to_ecg = hd5_path_to_ecg
        self._name = name
        self.transforms = transforms or []
        self.ecg_len = ecg_len
        self.leads = leads
        # s3
        self.local_only = True

    def _prep_s3_bucket_paths(
            self,
            s3_bucket_paths: Optional[Union[str, List[str]]],
    ) -> List[str]:
        """Prepares user input s3 paths for internal use"""
        if s3_bucket_paths is None:
            return
        elif isinstance(s3_bucket_paths, list):
            paths = s3_bucket_paths
        elif isinstance(s3_bucket_paths, str):
            paths = [s3_bucket_paths]
        else:
            raise TypeError(
                f"Cannot use s3_bucket_path input of type {type(s3_bucket_paths)}.",
            )
        for path in paths:
            os.makedirs(self._local_path(path), exist_ok=True)
        return paths

    def _local_path(self) -> str:
        """The folder downloaded hd5s from s3_bucket_path end up in"""
        if self.local_only:
            return self.local_hd5_folder

    def local_sample_ids(self) -> int:
        yield from map(
            sample_id_from_path,
            glob.glob(os.path.join(self.local_hd5_folder, "*.hd5")),
        )

    def _loading_options(
            self,
            sample_id: int,
            local_hd5_folder: str,
    ) -> List[LoadingOption]:
        hd5_path = os.path.join(local_hd5_folder, f"{sample_id}.hd5")
        with h5py.File(hd5_path, "r") as hd5:
            dates = list(
                hd5[self.hd5_path_to_ecg],
            )  # list all of the dates of saved ECGs
            sites = [
                decompress_data(
                    data_compressed=hd5[f'{self.hd5_path_to_ecg}/{date}/sitename'][()],
                    dtype='str'
                )
                for date in dates
            ]

        # print(f'got sites {sites} and dates {dates}')

        return [
            {
                'SITE': site,
                DATE_OPTION_KEY: pd.to_datetime(date).to_pydatetime(),
            }
            for date, site in zip(dates, sites)
        ]

    def get_loading_options(self, sample_id) -> List[LoadingOption]:
        options = []
        if self.local_only:
            return self._loading_options(sample_id, self.local_hd5_folder)
        return options

    def get_raw_data(self, sample_id, loading_option: LoadingOption) -> Tensor:
        hd5_path = os.path.join(self._local_path(), f"{sample_id}.hd5")
        with h5py.File(hd5_path, "r") as hd5:
            # we have to use the date provided in the loading option to get data from the hd5
            date_str = loading_option[DATE_OPTION_KEY].strftime(
                self.date_format,
            )

            # use the date to navigate to leads of the ECG
            compressed_leads = {
                lead: hd5[self.hd5_path_to_ecg][date_str][lead] for lead in self.leads
            }

            # decompress the ECG once retrieved
            ecg = np.zeros((self.ecg_len, len(self.leads)), dtype=np.float32)
            for lead in compressed_leads:
                voltages = decompress_data(
                    compressed_leads[lead][()],
                    dtype=compressed_leads[lead].attrs["dtype"],
                )
                if voltages.shape[0] != self.ecg_len:
                    voltages = np.interp(
                        np.linspace(0, 1, self.ecg_len),
                        np.linspace(0, 1, voltages.shape[0]),
                        voltages,
                    )
                ecg[:, self.leads[lead]] = voltages
            for transform in self.transforms:
                ecg = transform(ecg, loading_option)
            return ecg

    @property
    def name(self):
        return self._name

    def process_read(self, read: str) -> Dict[str, bool]:
        # TODO: speed this up, does a for loop over read for every kw
        return {
            name: any(kw in read for kw in kws)
            for name, kws in self.TEXT_DIAGNOSES.items()
        }

    def get_meta_data(self, hd5: h5py.Dataset) -> Dict[str, Any]:
        """hd5 should be keyed to desired ECG date"""
        out = {}
        for key in self.META_DATA_FIELDS:
            val = np.nan
            if key in hd5:
                dset = hd5[key]
                val = decompress_data(dset[()], dset.attrs["dtype"])
            out[key] = val
        return out

    def get_summary_data(self, sample_id, loading_option):
        """
        This function lets us pair down the 5000 x 12 data into something manageable
        for a DataFrame built by data exploration functions
        """
        # ECG meta data
        hd5_path = os.path.join(self._local_path(), f"{sample_id}.hd5")
        with h5py.File(hd5_path, "r") as hd5:
            date_str = loading_option[DATE_OPTION_KEY].strftime(
                self.date_format,
            )
            # get meta data
            out = self.get_meta_data(hd5[self.hd5_path_to_ecg][date_str])
            if "read_md_clean" in hd5[self.hd5_path_to_ecg][date_str]:
                read = hd5[self.hd5_path_to_ecg][date_str]["read_md_clean"]
                read = decompress_data(read[()], read.attrs["dtype"]).lower()
                out["read"] = "md"

            elif "read_pc_clean" in hd5[self.hd5_path_to_ecg][date_str]:
                read = hd5[self.hd5_path_to_ecg][date_str]["read_pc_clean"]
                read = decompress_data(read[()], read.attrs["dtype"]).lower()
                out["read"] = "pc"
            else:
                read = ""
                out["read"] = "none"

        # Raw ECG information
        ecg = self.get_raw_data(sample_id, loading_option)
        out["max_absolute_amp"] = np.abs(ecg).max()
        out["num_zeros"] = np.count_nonzero(ecg == 0)

        for lead_name, lead_idx in ECG_REST_AMP_LEADS.items():
            out[f"{lead_name}_zeros"] = np.count_nonzero(ecg[..., lead_idx] == 0)
            out = {**out, **self.process_read(read)}
        return out


class DataFrameDataDescription(DataDescription):
    def __init__(
            self,
            df: pd.DataFrame,
            col: str,
            process_col: Callable[[Any], Tensor] = None,
            name: str = None,
    ):
        """
        Gets data from a column of the provided DataFrame.
        :param df: Must be multi-indexed with sample_id, loading_option
        # TODO: allow multiple loading options
        :param col: The column name to get data from
        :param process_col: Function to turn the column value into Tensor
        """
        self.process_col = process_col or self._default_process_call
        self.df = df
        self.col = col
        self._name = name or col

    @staticmethod
    def _default_process_call(x: Any) -> Tensor:
        return np.array(x)

    @property
    def name(self) -> str:
        return self._name

    def get_loading_options(self, sample_id):
        row = self.df.loc[sample_id]
        return [{
            'start_fu_datetime': pd.to_datetime(row['start_fu_datetime']),
        }]

    def get_raw_data(
            self,
            sample_id: SampleID,
            loading_option: LoadingOption,
    ) -> Tensor:
        col_val = self.df.loc[sample_id][self.col]
        if self.col == 'age_in_days' and 'day_delta' in loading_option:
            col_val -= loading_option['day_delta']
        return self.process_col(col_val)


def one_hot_sex(x):
    return np.array([1, 0], dtype=np.float32) if x in [0, "Female"] else np.array([0, 1], dtype=np.float32)


def make_zscore(mu, std):
    def zscore(x):
        return (x-mu) / (1e-8+std)
    return zscore


def dataframe_data_description_from_tensor_map(
        tensor_map: TensorMap,
        dataframe: pd.DataFrame,
        is_input: bool = False,
) -> DataDescription:
    if tensor_map.is_categorical():
        process_col = one_hot_sex
    else:
        process_col = make_zscore(dataframe[tensor_map.name].mean(), dataframe[tensor_map.name].std())
    return DataFrameDataDescription(
        dataframe,
        col = tensor_map.name,
        process_col = process_col,
        name = tensor_map.input_name() if is_input else tensor_map.output_name(),
    )


class SurvivalWideFile(DataDescription):
    # DataDescription for a wide file

    def __init__(
            self,
            name: str,
            wide_df: pd.DataFrame,
            intervals: int = 25,
            follow_up_years: int = 10,
            event_age: str = 'hf_nlp_age',
            event_column: str = 'hf_nlp_event',
            ecg_age_column: str = 'ecg_age',
            start_age_column: str = 'start_fu_age',
            end_age_column: str = 'last_encounter',
    ):
        """
        """
        self.name = name
        self.intervals = intervals

        self.wide_df = wide_df
        self.event_age = event_age
        self.event_column = event_column
        self.ecg_age_column = ecg_age_column
        self.end_age_column = end_age_column
        self.follow_up_years = follow_up_years
        self.start_age_column = start_age_column
        print(f'Survival Curve with {365.25 * self.follow_up_years / self.intervals:.1f} days per interval.')

    def get_loading_options(self, sample_id):
        row = self.wide_df.loc[sample_id]
        ecg_date = pd.to_datetime(row[DATE_OPTION_KEY])
        start_date = ecg_date + (
                    pd.to_timedelta(row[self.start_age_column]) - pd.to_timedelta(row[self.ecg_age_column]))
        return [{
            DATE_OPTION_KEY: ecg_date,
            'start_date': start_date,
            'start_age': row[self.start_age_column],
            'event_age': row[self.event_age],
            'event': row[self.event_column],
        }]

    def get_raw_data(self, sample_id, loading_option):
        """expects time of ECG in the loading option as DATE_OPTION_KEY"""
        ecg_date = loading_option[DATE_OPTION_KEY]
        start_date = loading_option['start_date']
        ecg_age_difference = ecg_date - start_date

        row = self.wide_df.loc[sample_id]
        ecg_age = row[self.start_age_column] + ecg_age_difference

        has_disease = row[self.event_column]

        #         print(f"""
        #               OKAY: start_date {start_date} ecg_date {ecg_date} ecg_age_difference {ecg_age_difference}
        #               start age: {row[self.start_age_column]} ecg_age {ecg_age} has_disease: {has_disease}
        #               event age: {row[self.event_age]}, last encounter age: {row[self.end_age_column]}
        #               """)
        if has_disease:
            follow_up = pd.to_timedelta(row[self.event_age]) - ecg_age
        else:
            follow_up = pd.to_timedelta(row[self.end_age_column]) - ecg_age

        censor_date = ecg_date + follow_up
        days_per_interval = 365.25 * self.follow_up_years / self.intervals
        survival_then_censor = np.zeros(self.intervals * 2, dtype=np.float32)

        for i, day_delta in enumerate(np.arange(0, 365.25 * self.follow_up_years, days_per_interval)):
            cur_date = ecg_date + datetime.timedelta(days=day_delta)
            survival_then_censor[i] = float(cur_date < censor_date)
            survival_then_censor[self.intervals + i] = has_disease * float(
                censor_date <= cur_date < censor_date + datetime.timedelta(days=days_per_interval))
        # Handle prevalent diseases
        if has_disease and pd.to_timedelta(row[self.event_age]) <= pd.to_timedelta(ecg_age):
            survival_then_censor[self.intervals] = has_disease
        return survival_then_censor

    def name(self):
        return self.name



# VTK-specific tensormaps
import os
import h5py
import numpy as np
import vtk
import vtk.util.numpy_support
from ml4h.TensorMap import TensorMap, Interpretation
from ml4h.defines import MRI_SEGMENTED, MRI_LAX_SEGMENTED, MRI_FRAMES, MRI_PIXEL_WIDTH, MRI_PIXEL_HEIGHT, MRI_SLICE_THICKNESS, MRI_PATIENT_ORIENTATION, MRI_PATIENT_POSITION
from ml4h.tensormap.ukb.mri import mri_tensor_2d


def _mri_tensor_4d(hd5, name):
    """
    Returns MRI image tensors from HD5 as 4-D numpy arrays. Useful for raw SAX and LAX images and segmentations.
    """
    if isinstance(hd5[name], h5py.Group):
        nslices = len(hd5[name]) // MRI_FRAMES
        for img in hd5[name]:
            img_shape = hd5[name][img].shape
            break
        shape = (img_shape[0], img_shape[1], nslices, MRI_FRAMES)
        arr = np.zeros(shape)
        t = 0
        s = 0
        for k in sorted(hd5[name], key=int):
            arr[:, :, s, t] = np.array(hd5[name][k]).T
            t += 1
            if t == MRI_FRAMES:
                s += 1
                t = 0
    elif isinstance(hd5[name], h5py.Dataset):
        nslices = 1
        shape = (hd5[name].shape[0], hd5[name].shape[1], nslices, MRI_FRAMES)
        arr = np.zeros(shape)
        for t in range(MRI_FRAMES):
            arr[:, :, 0, t] = np.array(hd5[name][:, :, t]).T
    else:
        raise ValueError(f'{name} is neither a HD5 Group nor a HD5 dataset')
    return arr


def _mri_hd5_to_structured_grids(hd5, name, save_path=None, order='F'):
    """
    Returns MRI tensors as list of VTK structured grids aligned to the reference system of the patient
    """
    arr = _mri_tensor_4d(hd5, name)
    width = hd5['_'.join([MRI_PIXEL_WIDTH, name])]
    height = hd5['_'.join([MRI_PIXEL_HEIGHT, name])]
    positions = mri_tensor_2d(hd5, '_'.join([MRI_PATIENT_POSITION, name]))
    orientations = mri_tensor_2d(
        hd5,
        '_'.join([MRI_PATIENT_ORIENTATION, name]),
    )
    thickness = hd5['_'.join([MRI_SLICE_THICKNESS, name])]
    _, dataset_indices, dataset_counts = np.unique(
        orientations,
        axis=1,
        return_index=True,
        return_counts=True,
    )
    grids = []
    for d_idx, d_cnt in zip(dataset_indices, dataset_counts):
        grids.append(vtk.vtkStructuredGrid())
        nslices = d_cnt
        # If multislice, override thickness as distance between voxel centers. Note: removes eventual gaps between slices
        if nslices > 1:
            thickness = np.linalg.norm(
                positions[:, d_idx] -
                positions[:, d_idx + 1],
            )
        transform = vtk.vtkTransform()
        n_orientation = np.cross(
            orientations[3:, d_idx], orientations[
                :3,
                d_idx,
            ],
        )
        # 4x4 transform matrix to align to the patient reference system
        transform.SetMatrix([
            orientations[3, d_idx] * height,
            orientations[0, d_idx] * width,
            n_orientation[0] * thickness,
            positions[0, d_idx],
            orientations[4, d_idx] * height,
            orientations[1, d_idx] * width,
            n_orientation[1] * thickness,
            positions[1, d_idx],
            orientations[5, d_idx] * height,
            orientations[2, d_idx] * width,
            n_orientation[2] * thickness,
            positions[2, d_idx],
            0,
            0,
            0,
            1,
        ])
        x_coors = np.arange(0, arr.shape[0] + 1) - 0.5
        y_coors = np.arange(0, arr.shape[1] + 1) - 0.5
        z_coors = np.arange(0, d_cnt + 1) - 0.5
        xyz_meshgrid = np.meshgrid(x_coors, y_coors, z_coors)
        xyz_pts = np.zeros(
            ((arr.shape[0] + 1) * (arr.shape[1] + 1) * (d_cnt + 1), 3),
        )
        for dim in range(3):
            xyz_pts[:, dim] = xyz_meshgrid[dim].ravel(order=order)
        vtk_pts = vtk.vtkPoints()
        vtk_pts.SetData(vtk.util.numpy_support.numpy_to_vtk(xyz_pts))
        grids[-1].SetPoints(vtk_pts)
        grids[-1].SetDimensions(len(x_coors), len(y_coors), len(z_coors))
        grids[-1].SetExtent(
            0,
            len(x_coors) - 1, 0,
            len(y_coors) - 1, 0,
            len(z_coors) - 1,
        )
        for t in range(MRI_FRAMES):
            arr_vtk = vtk.util.numpy_support.numpy_to_vtk(
                arr[:, :, d_idx:d_idx + d_cnt, t].ravel(order=order),
                deep=True,
            )
            arr_vtk.SetName(f'{name}_{t}')
            grids[-1].GetCellData().AddArray(arr_vtk)
        transform_filter = vtk.vtkTransformFilter()
        transform_filter.SetInputData(grids[-1])
        transform_filter.SetTransform(transform)
        transform_filter.Update()
        grids[-1].DeepCopy(transform_filter.GetOutput())
        if save_path:
            writer = vtk.vtkXMLStructuredGridWriter()
            writer.SetFileName(
                os.path.join(save_path, f'grid_{name}_{d_idx}.vts'),
            )
            writer.SetInputData(grids[-1])
            writer.Update()
    return grids


def _cut_through_plane(dataset, plane_center, plane_orientation):
    plane = vtk.vtkPlane()
    plane.SetOrigin(plane_center)
    plane.SetNormal(plane_orientation)
    cutter = vtk.vtkCutter()
    cutter.SetInputData(dataset)
    cutter.SetCutFunction(plane)
    poly = vtk.vtkDataSetSurfaceFilter()
    poly.SetInputConnection(cutter.GetOutputPort())
    poly.Update()
    return poly.GetOutput()


def _map_points_to_cells(pts, dataset, tol=1e-3):
    locator = vtk.vtkCellLocator()
    locator.SetDataSet(dataset)
    locator.BuildLocator()
    closest_pt = np.zeros(3)
    generic_cell = vtk.vtkGenericCell()
    cell_id, sub_id, dist2, inside = vtk.mutable(0), vtk.mutable(
        0,
    ), vtk.mutable(0.0), vtk.mutable(0)
    map_to_cells = np.zeros(len(pts), dtype=np.int64)
    for pt_id, pt in enumerate(pts):
        if locator.FindClosestPointWithinRadius(
            pt, tol, closest_pt,
            generic_cell, cell_id, sub_id,
            dist2, inside,
        ):
            map_to_cells[pt_id] = cell_id.get()
    return map_to_cells


def _make_mri_projected_segmentation_from_file(
    to_segment_name,
    segmented_name,
    save_path=None,
):
    def mri_projected_segmentation(tm, hd5):
        if segmented_name not in [MRI_SEGMENTED, MRI_LAX_SEGMENTED]:
            raise ValueError(
                f'{segmented_name} is recognized neither as SAX nor LAX segmentation',
            )
        cine_segmented_grids = _mri_hd5_to_structured_grids(
            hd5, segmented_name,
        )
        cine_to_segment_grids = _mri_hd5_to_structured_grids(
            hd5, to_segment_name,
        )
        tensor = np.zeros(tm.shape, dtype=np.float32)
        # Loop through segmentations and datasets
        for ds_i, ds_segmented in enumerate(cine_segmented_grids):
            for ds_j, ds_to_segment in enumerate(cine_to_segment_grids):
                dims = ds_to_segment.GetDimensions()
                pts = vtk.util.numpy_support.vtk_to_numpy(
                    ds_to_segment.GetPoints().GetData(),
                )
                npts_per_slice = dims[0] * dims[1]
                ncells_per_slice = (dims[0] - 1) * (dims[1] - 1)
                n_orientation = (pts[npts_per_slice] - pts[0])
                n_orientation /= np.linalg.norm(n_orientation)
                cell_centers = vtk.vtkCellCenters()
                cell_centers.SetInputData(ds_to_segment)
                cell_centers.Update()
                cell_pts = vtk.util.numpy_support.vtk_to_numpy(
                    cell_centers.GetOutput().GetPoints().GetData(),
                )
                # Loop through dataset slices
                for s in range(dims[2] - 1):
                    slice_center = np.mean(
                        pts[
                            s * npts_per_slice:(s + 2) *
                            npts_per_slice
                        ],
                        axis=0,
                    )
                    slice_cell_pts = cell_pts[
                        s * ncells_per_slice:(s + 1) *
                        ncells_per_slice
                    ]
                    slice_segmented = _cut_through_plane(
                        ds_segmented, slice_center, n_orientation,
                    )
                    map_to_segmented = _map_points_to_cells(
                        slice_cell_pts, slice_segmented,
                    )
                    # Loop through time
                    for t in range(MRI_FRAMES):
                        arr_name = f'{segmented_name}_{t}'
                        segmented_arr = vtk.util.numpy_support.vtk_to_numpy(
                            slice_segmented.GetCellData().GetArray(arr_name),
                        )
                        projected_arr = segmented_arr[map_to_segmented]
                        if len(tm.shape) == 3:
                            tensor[:, :, t] = np.maximum(
                                tensor[:, :, t],
                                projected_arr.reshape(
                                    tm.shape[0],
                                    tm.shape[1],
                                ),
                            )
                        elif len(tm.shape) == 4:
                            tensor[:, :, s, t] = np.maximum(
                                tensor[:, :, s, t],
                                projected_arr.reshape(
                                    tm.shape[0],
                                    tm.shape[1],
                                ),
                            )
                    if save_path:
                        writer_segmented = vtk.vtkXMLPolyDataWriter()
                        writer_segmented.SetInputData(slice_segmented)
                        writer_segmented.SetFileName(
                            os.path.join(
                                save_path,
                                f'{tm.name}_segmented_{ds_i}_{ds_j}_{s}.vtp',
                            ),
                        )
                        writer_segmented.Update()
        return tensor

    return mri_projected_segmentation


cine_segmented_lax_2ch_proj_from_sax = TensorMap(
    'cine_segmented_lax_2ch_proj_from_sax',
    Interpretation.CONTINUOUS,
    shape=(256, 256, 50),
    loss='logcosh',
    tensor_from_file=_make_mri_projected_segmentation_from_file(
        'cine_segmented_lax_2ch', MRI_SEGMENTED,
    ),
)
cine_segmented_lax_3ch_proj_from_sax = TensorMap(
    'cine_segmented_lax_3ch_proj_from_sax',
    Interpretation.CONTINUOUS,
    shape=(256, 256, 50),
    loss='logcosh',
    tensor_from_file=_make_mri_projected_segmentation_from_file(
        'cine_segmented_lax_3ch', MRI_SEGMENTED,
    ),
)
cine_segmented_lax_4ch_proj_from_sax = TensorMap(
    'cine_segmented_lax_4ch_proj_from_sax',
    Interpretation.CONTINUOUS,
    shape=(256, 256, 50),
    loss='logcosh',
    tensor_from_file=_make_mri_projected_segmentation_from_file(
        'cine_segmented_lax_4ch', MRI_SEGMENTED,
    ),
)
cine_segmented_lax_2ch_proj_from_lax = TensorMap(
    'cine_segmented_lax_2ch_proj_from_lax',
    Interpretation.CONTINUOUS,
    shape=(256, 256, 50),
    loss='logcosh',
    tensor_from_file=_make_mri_projected_segmentation_from_file(
        'cine_segmented_lax_2ch', MRI_LAX_SEGMENTED,
    ),
)
cine_segmented_lax_3ch_proj_from_lax = TensorMap(
    'cine_segmented_lax_3ch_proj_from_lax',
    Interpretation.CONTINUOUS,
    shape=(256, 256, 50),
    loss='logcosh',
    tensor_from_file=_make_mri_projected_segmentation_from_file(
        'cine_segmented_lax_3ch', MRI_LAX_SEGMENTED,
    ),
)
cine_segmented_lax_4ch_proj_from_lax = TensorMap(
    'cine_segmented_lax_4ch_proj_from_lax',
    Interpretation.CONTINUOUS,
    shape=(256, 256, 50),
    loss='logcosh',
    tensor_from_file=_make_mri_projected_segmentation_from_file(
        'cine_segmented_lax_4ch', MRI_LAX_SEGMENTED,
    ),
)



"""Methods for integration of interactive DICOM plots within notebooks.

TODO:
* Continue to *pragmatically* improve this to make the visualization controls
  and the information shown more similar to that in full-featured DICOM viewers.
"""

import collections
import os
import tempfile
from typing import Any, DefaultDict, Dict, Optional, Tuple
import zipfile

from IPython.display import display
from IPython.display import HTML
import numpy as np
import ipywidgets as widgets
import matplotlib.pyplot as plt
from ml4h.runtime_data_defines import get_mri_folders
import pydicom
import tensorflow as tf

MIN_IMAGE_WIDTH = 8
DEFAULT_IMAGE_WIDTH = 12
MAX_IMAGE_WIDTH = 24

MIN_COLOR_RANGE = 0
MAX_COLOR_RANGE = 6000


def choose_mri(sample_id, folder: Optional[str] = None) -> None:
  """Render widget to choose the MRI to plot.

  Args:
    sample_id: The id of the sample to retrieve.
    folder: The local or Cloud Storage folder under which the files reside.
  """
  if folder is None:
    folders = get_mri_folders(sample_id)
  else:
    folders = [folder]

  sample_mris = []
  sample_mri_glob = str(sample_id) + '_*.zip'
  try:
    for f in folders:
      sample_mris.extend(tf.io.gfile.glob(pattern=os.path.join(f, sample_mri_glob)))
  except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
    display(
        HTML(f'''<div class="alert alert-block alert-danger">
    <b>Warning:</b> MRI not available for sample {sample_id} in {folders}:
    <hr><p><pre>{e.message}</pre></p>
    Use the <kbd>folder</kbd> parameter to read DICOMs from a different local directory or Cloud Storage bucket.
    </div>'''),
    )
    return

  if not sample_mris:
    display(
        HTML(f'''<div class="alert alert-block alert-danger">
    <b>Warning:</b> MRI DICOMs not available for sample {sample_id} in {folders}.<br>
    Use the <kbd>folder</kbd> parameter to read DICOMs from a different local directory or Cloud Storage bucket.
    </div>'''),
    )
    return

  mri_chooser = widgets.Dropdown(
      options=sample_mris,
      value=sample_mris[0],
      description=f'Choose an MRI to visualize for sample {sample_id}:',
      style={'description_width': 'initial'},
      layout=widgets.Layout(width='800px'),
  )
  file_controls_ui = widgets.VBox(
      [widgets.HTML('<h3>File controls</h3>'), mri_chooser],
      layout=widgets.Layout(width='auto', border='solid 1px grey'),
  )
  file_controls_output = widgets.interactive_output(choose_mri_series, {'sample_mri': mri_chooser})
  display(file_controls_ui, file_controls_output)


def choose_mri_series(sample_mri: str) -> None:
  """Render widgets and interactive plots for MRIs.

  Args:
    sample_mri: The local or Cloud Storage path to the MRI file.
  """
  with tempfile.TemporaryDirectory() as tmpdirname:
    local_path = os.path.join(tmpdirname, os.path.basename(sample_mri))
    try:
      tf.io.gfile.copy(src=sample_mri, dst=local_path)
      with zipfile.ZipFile(local_path, 'r') as zip_ref:
        zip_ref.extractall(tmpdirname)
    except (tf.errors.NotFoundError, tf.errors.PermissionDeniedError) as e:
      display(
          HTML(f'''<div class="alert alert-block alert-danger">
      <b>Warning:</b> Cardiac MRI not available for sample {os.path.basename(sample_mri)}:
      <hr><p><pre>{e.message}</pre></p>
      </div>'''),
      )
      return

    unordered_dicoms: DefaultDict[Any, Any] = collections.defaultdict(dict)
    for dcm_file in os.listdir(tmpdirname):
      if not dcm_file.endswith('.dcm'):
        continue
      dcm = pydicom.read_file(os.path.join(tmpdirname, dcm_file))
      key1 = (dcm.SeriesDescription.lower(), int(dcm.SeriesNumber))
      key2 = int(dcm.InstanceNumber) - 1
      if key2 in unordered_dicoms[key1]:
        # Notice invalid input, but don't throw an error.
        print(f'WARNING: Duplicate instances: {dcm.SeriesDescription} {dcm.SeriesNumber} {dcm.InstanceNumber}.')
      unordered_dicoms[key1][key2] = dcm

  if not unordered_dicoms:
    display(
        HTML(f'''<div class="alert alert-block alert-warning">
    No series available in MRI for sample {os.path.basename(sample_mri)}.
    Try a different MRI.
    </div>'''),
    )
    return

  # Convert from dict of dicts to dict of ordered lists.
  dicoms = {}
  for series in unordered_dicoms.keys():
    dicoms[series] = [None] * (max(unordered_dicoms[series]) + 1)
    for idx, val in unordered_dicoms[series].items():
      dicoms[series][idx] = val

  default_series_value = sorted(list(dicoms.keys()))[0]
  # Display the middle instance by default.
  default_instance_value, max_instance_value = compute_instance_range(dicoms, default_series_value)
  default_vmin_value, default_vmax_value = compute_color_range(dicoms, default_series_value)

  series_name_chooser = widgets.Dropdown(
      options=[(str(k), k) for k in sorted(dicoms.keys())],
      value=default_series_value,
      description='Choose the MRI series to visualize:',
      style={'description_width': 'initial'},
      layout=widgets.Layout(width='800px'),
  )
  # Slide through DICOM image instances using a slide bar.
  instance_chooser = widgets.IntSlider(
      continuous_update=True,
      value=default_instance_value,
      min=1,
      max=max_instance_value,
      description='Image instance to display '
      + '(click on slider, then use left/right arrows):',
      style={'description_width': 'initial'},
      layout=series_name_chooser.layout,
  )
  vmin_chooser = widgets.IntSlider(
      continuous_update=True,
      value=default_vmin_value,
      min=MIN_COLOR_RANGE,
      max=MAX_COLOR_RANGE,
      description='Color range minimum:',
      style={'description_width': 'initial'},
      layout=widgets.Layout(width='300px'),
  )
  vmax_chooser = widgets.IntSlider(
      continuous_update=True,
      value=default_vmax_value,
      min=MIN_COLOR_RANGE,
      max=MAX_COLOR_RANGE,
      description='Color range maximum:',
      style={'description_width': 'initial'},
      layout=vmin_chooser.layout,
  )
  transpose_chooser = widgets.Checkbox(
      description='Whether to transpose the image.',
      style={'description_width': 'initial'},
      layout=vmin_chooser.layout,
  )
  fig_width_chooser = widgets.IntSlider(
      continuous_update=False,
      value=DEFAULT_IMAGE_WIDTH,
      min=MIN_IMAGE_WIDTH,
      max=MAX_IMAGE_WIDTH,
      description='Width of figure (height will be computed using input data):',
      style={'description_width': 'initial'},
      layout=vmin_chooser.layout,
  )

  viz_controls_ui = widgets.VBox(
      [
          widgets.HTML('<h3>Visualization controls</h3>'),
          series_name_chooser, instance_chooser,
          widgets.HBox([vmin_chooser, vmax_chooser]),
          widgets.HBox([transpose_chooser, fig_width_chooser]),
      ],
      layout=widgets.Layout(width='auto', border='solid 1px grey'),
  )
  viz_controls_output = widgets.interactive_output(
      dicom_animation,
      {
          'dicoms': widgets.fixed(dicoms),
          'series_name': series_name_chooser,
          'instance': instance_chooser,
          'vmin': vmin_chooser,
          'vmax': vmax_chooser,
          'transpose': transpose_chooser,
          'fig_width': fig_width_chooser,
          'title_prefix': widgets.fixed(os.path.basename(sample_mri)),
      },
  )

  def on_value_change(change):
    """Inner function to capture state being observed."""
    vmin_chooser.value, vmax_chooser.value = compute_color_range(dicoms, change['new'])
    instance_chooser.value, instance_chooser.max = compute_instance_range(dicoms, change['new'])

  # When the series changes, update the widgets to the proper ranges
  # for the series.
  series_name_chooser.observe(on_value_change, names='value')
  display(viz_controls_ui, viz_controls_output)


def compute_color_range(dicoms: Dict[str, Any], series_name: str) -> Tuple[int, int]:
  """Compute the mean values for the color ranges of instances in the series."""
  vmin = np.mean([np.min(d.pixel_array) for d in dicoms[series_name]])
  vmax = np.mean([np.max(d.pixel_array) for d in dicoms[series_name]])
  return (vmin, vmax)


def compute_instance_range(dicoms: Dict[str, Any], series_name: str) -> Tuple[int, int]:
  """Compute middle and max instances."""
  middle_instance = int(len(dicoms[series_name]) / 2)
  max_instance = len(dicoms[series_name])
  return (middle_instance, max_instance)


def dicom_animation(
    dicoms: Dict[str, Any], series_name: str, instance: int, vmin: int, vmax: int, transpose: bool,
    fig_width: int, title_prefix: str = '',
) -> None:
  """Render one frame of a DICOM animation.

  Args:
    dicoms: the dictionary DICOM series and instances lists
    series_name: the name of the series to be displayed
    instance: the particular instance to display
    vmin: minimum value for the color range
    vmax: maximum value for the color range
    transpose: whether or not to transpose the image
    fig_width: the desired width of the figure, note that height computed as
      the proportion of the width based on the data to be plotted
    title_prefix: text to display as the initial portion of the plot title
  """
  if len(dicoms[series_name]) < instance:
    dcm = dicoms[series_name][-1]
    print(f'Instance {str(instance)} not available for {series_name}, using final instance instead.')
  else:
    dcm = dicoms[series_name][instance - 1]
    if instance != dcm.InstanceNumber:
      # Notice invalid input, but don't throw an error.
      print(f'WARNING: Instance parameter {str(instance)} and instance number {str(dcm.InstanceNumber)} do not match.')

  if transpose:
    height = dcm.pixel_array.T.shape[0]
    width = dcm.pixel_array.T.shape[1]
  else:
    height = dcm.pixel_array.shape[0]
    width = dcm.pixel_array.shape[1]

  fig_height = int(np.ceil(fig_width * (height/width)))

  _, ax = plt.subplots(figsize=(fig_width, fig_height), facecolor='beige')
  ax.imshow(dcm.pixel_array.T if transpose else dcm.pixel_array, cmap='gray', vmin=vmin, vmax=vmax)
  ax.set_title(
      title_prefix
      + ', Series: ' + dcm.SeriesDescription
      + ', Series Number: ' + str(dcm.SeriesNumber)
      + ', Instance: ' + str(dcm.InstanceNumber)
      + '\nColor range: ' + str(vmin) + '-' + str(vmax)
      + ', Transpose: ' + str(transpose)
      + ', Figure size:' + str(fig_width) + 'x' + str(fig_height),
      fontsize=fig_width,
  )
  ax.set_yticklabels([])
  ax.set_xticklabels([])



import h5py
import numpy as np
import scipy
import logging

import biosppy
from typing import List, Tuple, Dict
from tensorflow.keras.utils import to_categorical

from ml4h.tensormap.ukb.demographics import age_in_years_tensor
from ml4h.normalizer import ZeroMeanStd1, Standardize, RandomStandardize
from ml4h.metrics import weighted_crossentropy, ignore_zeros_logcosh, mse_10x
from ml4h.TensorMap import TensorMap, Interpretation, no_nans, make_range_validator
from ml4h.defines import ECG_CHAR_2_IDX, ECG_REST_MGB_LEADS, ECG_REST_AMP_LEADS_UKB, ECG_BIKE_LEADS
from ml4h.defines import ECG_REST_LEADS, ECG_REST_MEDIAN_LEADS, ECG_REST_AMP_LEADS, ECG_SEGMENTED_CHANNEL_MAP
from ml4h.tensormap.general import tensor_path, pad_or_crop_array_to_shape, tensor_from_hd5, named_tensor_from_hd5
from ml4h.tensormap.general import get_tensor_at_first_date, normalized_first_date, pass_nan, build_tensor_from_file

_HRR_SENTINEL = -1000


# BIKE ECG
def _check_phase_full_len(hd5: h5py.File, phase: str):
    phase_len = get_tensor_at_first_date(hd5, 'ecg_bike', f'{phase}_duration')
    valid = True
    if phase == 'pretest':
        valid &= phase_len == 15
    elif phase == 'exercise':
        valid &= phase_len == 360
    elif phase == 'rest':
        valid &= phase_len == 60
    else:
        raise ValueError(f'Phase {phase} is not a valid phase.')
    if not valid:
        raise ValueError(f'{phase} phase is not full length')


def _first_date_bike_recovery(tm: TensorMap, hd5: h5py.File, dependents=None):
    _check_phase_full_len(hd5, 'rest')
    original = get_tensor_at_first_date(hd5, tm.path_prefix, tm.name)
    recovery = original[-tm.shape[0]:]
    return recovery.reshape(tm.shape)


def _first_date_bike_pretest(tm: TensorMap, hd5: h5py.File, dependents=None):
    _check_phase_full_len(hd5, 'pretest')
    original = get_tensor_at_first_date(hd5, tm.path_prefix, tm.name)
    pretest = original[:tm.shape[0]]
    return pretest.reshape(tm.shape)


def _first_date_hrr(tm: TensorMap, hd5: h5py.File, dependents=None):
    _check_phase_full_len(hd5, 'rest')
    last_hr = get_tensor_at_first_date(hd5, 'ecg_bike', 'trend_heartrate')[-1]
    max_hr = get_tensor_at_first_date(hd5, 'ecg_bike', 'max_hr')
    return max_hr - last_hr


def _healthy_check(hd5):
    for phase in ('pretest', 'exercise', 'rest'):
        _check_phase_full_len(hd5, phase)
    max_load = max(get_tensor_at_first_date(hd5, 'ecg_bike', 'trend_load'))
    if max_load < 60:
        raise ValueError('Max load not high enough')


def _healthy_bike(tm: TensorMap, hd5: h5py.File, dependents=None):
    _healthy_check(hd5)
    return normalized_first_date(tm, hd5)


def _healthy_hrr(tm: TensorMap, hd5: h5py.File, dependents=None):
    _healthy_check(hd5)
    return _first_date_hrr(tm, hd5)


def _median_pretest(tm: TensorMap, hd5: h5py.File, dependents=None):
    _healthy_check(hd5)
    times = get_tensor_at_first_date(hd5, 'ecg_bike', 'trend_time')
    tensor = np.abs(
        get_tensor_at_first_date(
        hd5, tm.path_prefix, 'float_array', tm.name,
        ),
    )
    return np.median(tensor[times <= 15])


def _new_hrr(tm: TensorMap, hd5: h5py.File, dependents=None):
    _check_phase_full_len(hd5, 'rest')
    hrs = get_tensor_at_first_date(hd5, 'ecg_bike', 'trend_heartrate')
    phases = get_tensor_at_first_date(hd5, 'ecg_bike', 'trend_phasename')
    min_hr = hrs[phases == 2].min()
    max_hr = get_tensor_at_first_date(hd5, 'ecg_bike', 'max_hr')
    max_pred = get_tensor_at_first_date(hd5, 'ecg_bike', 'max_pred_hr')
    hrr = max_hr - min_hr
    if max_hr / max_pred > 150:
        raise ValueError('Max hr / max pred hr too high.')
    if hrr > 80:
        raise ValueError('HRR too high.')
    return hrr


def _sentinel_hrr(tm: TensorMap, hd5: h5py.File, dependents=None):
    try:
        return _new_hrr(tm, hd5)
    except ValueError:
        return _HRR_SENTINEL


def _hr_achieved(tm: TensorMap, hd5: h5py.File, dependents=None):
    _check_phase_full_len(hd5, 'rest')
    max_hr = get_tensor_at_first_date(hd5, 'ecg_bike', 'max_hr')
    max_pred = get_tensor_at_first_date(hd5, 'ecg_bike', 'max_pred_hr')
    return max_hr / max_pred


def _warp_ecg(ecg):
    i = np.arange(ecg.shape[0])
    warped = i + (
        np.random.rand() * 100 * np.sin(i / (500 + np.random.rand() * 100))
        + np.random.rand() * 100 * np.cos(i / (500 + np.random.rand() * 100))
    )
    warped_ecg = np.zeros_like(ecg)
    for j in range(ecg.shape[1]):
        warped_ecg[:, j] = np.interp(i, warped, ecg[:, j])
    return warped_ecg


def _make_ecg_rest(
        instance: int = 0, downsample_steps: int = 0,
        short_time_nperseg: int = 0, short_time_noverlap: int = 0,
        skip_poor: bool = False, random_offset: int = 0,
):
    def ecg_rest_from_file(tm, hd5, dependents={}):
        ecg_interpretation = str(
            tm.hd5_first_dataset_in_group(
            hd5, 'ukb_ecg_rest/ecg_rest_text/',
            )[()],
        )
        if skip_poor and 'Poor data quality' in ecg_interpretation:
            raise ValueError(f'Poor data quality skipped by {tm.name}.')
        tensor = np.zeros(tm.shape, dtype=np.float32)
        for k in hd5[tm.path_prefix]:
            if k in tm.channel_map:
                data = tm.hd5_first_dataset_in_group(
                    hd5, f'{tm.path_prefix}/{k}/instance_{instance}',
                )
                if random_offset > 0:
                    offset = np.random.randint(random_offset)
                    data = data[offset:]
                if short_time_nperseg > 0 and short_time_noverlap > 0:
                    f, t, short_time_ft = scipy.signal.stft(
                        data, nperseg=short_time_nperseg, noverlap=short_time_noverlap,
                    )
                    tensor[..., tm.channel_map[k]] = short_time_ft
                elif downsample_steps > 1:
                    tensor[:, tm.channel_map[k]] = np.array(data, dtype=np.float32)[::downsample_steps]
                else:
                    tensor[:, tm.channel_map[k]] = pad_or_crop_array_to_shape((tm.shape[0],), data)
        return tensor
    return ecg_rest_from_file


def _get_lead_cm(length):
    lead_cm = {}
    lead_weights = []
    for i in range(length):
        wave_val = i - (length//2)
        lead_cm['w'+str(wave_val).replace('-', '_')] = i
        lead_weights.append((np.abs(wave_val+1)/(length/2)) + 1.0)
    return lead_cm, lead_weights


def _make_rhythm_tensor(skip_poor=True):
    def rhythm_tensor_from_file(tm, hd5, dependents={}):
        categorical_data = np.zeros(tm.shape, dtype=np.float32)
        ecg_interpretation = str(
            tm.hd5_first_dataset_in_group(
            hd5, 'ukb_ecg_rest/ecg_rest_text/',
            )[()],
        )
        if skip_poor and 'Poor data quality' in ecg_interpretation:
            raise ValueError(f'Poor data quality skipped by {tm.name}.')
        for channel in tm.channel_map:
            if channel.replace('_', ' ') in ecg_interpretation:
                categorical_data[tm.channel_map[channel]] = 1.0
                return categorical_data
        for rhythm in ['sinus', 'Sinus']:
            if rhythm in ecg_interpretation:
                categorical_data[tm.channel_map['Other_sinus_rhythm']] = 1.0
                return categorical_data
        categorical_data[tm.channel_map['Other_rhythm']] = 1.0
        return categorical_data
    return rhythm_tensor_from_file


def label_from_ecg_interpretation_text(tm, hd5, dependents={}):
    categorical_data = np.zeros(tm.shape, dtype=np.float32)
    ecg_interpretation = str(
        tm.hd5_first_dataset_in_group(
        hd5, 'ukb_ecg_rest/ecg_rest_text/',
        )[()],
    )
    for channel in tm.channel_map:
        if channel.lower() in ecg_interpretation.lower():
            categorical_data[tm.channel_map[channel]] = 1.0
            return categorical_data
    if 'no_' + tm.name in tm.channel_map:
        categorical_data[tm.channel_map['no_' + tm.name]] = 1.0
        return categorical_data
    else:
        raise ValueError(
            f"ECG categorical interpretation could not find any of these keys: {tm.channel_map.keys()}",
        )


# Extract RAmplitude and SAmplitude for LVH criteria
def _make_ukb_ecg_rest(population_normalize: float = None):
    def ukb_ecg_rest_from_file(tm, hd5, dependents={}):
        if 'ukb_ecg_rest' not in hd5:
            raise ValueError(
                'Group with R and S amplitudes not present in hd5',
            )
        tensor = get_tensor_at_first_date(
            hd5, tm.path_prefix, tm.name, pass_nan,
        )
        try:
            if population_normalize is None:
                tensor = tm.zero_mean_std1(tensor)
            else:
                tensor /= population_normalize
        except:
            ValueError(f'Cannot normalize {tm.name}')
        return tensor
    return ukb_ecg_rest_from_file


def _make_ukb_ecg_rest_lvh():
    def ukb_ecg_rest_lvh_from_file(tm, hd5, dependents={}):
        # Lead order seems constant and standard throughout, but we could eventually tensorize it from XML
        lead_order = ECG_REST_AMP_LEADS
        avl_min = 1100.0
        sl_min = 3500.0
        cornell_female_min = 2000.0
        cornell_male_min = 2800.0
        if 'ukb_ecg_rest' not in hd5:
            raise ValueError(
                'Group with R and S amplitudes not present in hd5',
            )
        tensor_ramp = get_tensor_at_first_date(
            hd5, tm.path_prefix, 'ramplitude', pass_nan,
        )
        tensor_samp = get_tensor_at_first_date(
            hd5, tm.path_prefix, 'samplitude', pass_nan,
        )
        criteria_sleads = [lead_order[l] for l in ['V1', 'V3']]
        criteria_rleads = [lead_order[l] for l in ['aVL', 'V5', 'V6']]
        if np.any(np.isnan(np.union1d(tensor_ramp[criteria_rleads], tensor_samp[criteria_sleads]))):
            raise ValueError(
                'Missing some of the R and S amplitude readings needed to evaluate LVH criteria',
            )
        is_female = 'Genetic-sex_Female_0_0' in hd5['categorical']
        is_male = 'Genetic-sex_Male_0_0' in hd5['categorical']
        # If genetic sex not available, try phenotypic
        if not(is_female or is_male):
            is_female = 'Sex_Female_0_0' in hd5['categorical']
            is_male = 'Sex_Male_0_0' in hd5['categorical']
        # If neither available, raise error
        if not(is_female or is_male):
            raise ValueError('Sex info required to evaluate LVH criteria')
        if tm.name == 'avl_lvh':
            is_lvh = tensor_ramp[lead_order['aVL']] > avl_min
        elif tm.name == 'sokolow_lyon_lvh':
            is_lvh = tensor_samp[lead_order['V1']] +\
                np.maximum(tensor_ramp[lead_order['V5']], tensor_ramp[lead_order['V6']]) > sl_min
        elif tm.name == 'cornell_lvh':
            is_lvh = tensor_ramp[lead_order['aVL']] + \
                tensor_samp[lead_order['V3']]
            if is_female:
                is_lvh = is_lvh > cornell_female_min
            if is_male:
                is_lvh = is_lvh > cornell_male_min
        else:
            raise ValueError(
                f'{tm.name} criterion for LVH is not accounted for',
            )
        # Following convention from categorical TMAPS, positive has cmap index 1
        tensor = np.zeros(tm.shape, dtype=np.float32)
        index = 0
        if is_lvh:
            index = 1
        tensor[index] = 1.0
        return tensor
    return ukb_ecg_rest_lvh_from_file


def _ecg_rest_to_segment(population_normalize=None, hertz=500, random_offset_seconds=0):
    def ecg_rest_section_to_segment(tm, hd5, dependents={}):
        tensor = np.zeros(tm.shape, dtype=np.float32)
        segmented = tm.dependent_map.hd5_first_dataset_in_group(
            hd5, tm.dependent_map.hd5_key_guess(),
        )
        offset_seconds = float(segmented.attrs['offset_seconds'])
        random_offset_samples = 0
        if random_offset_seconds > 0:
            random_offset_begin = np.random.uniform(random_offset_seconds)
            offset_seconds += random_offset_begin
            random_offset_samples = int(random_offset_begin * hertz)
        offset_begin = int(offset_seconds * hertz)
        segment_index = np.array(
            segmented[random_offset_samples:random_offset_samples+tm.dependent_map.shape[0]], dtype=np.float32,
        )
        dependents[tm.dependent_map] = to_categorical(
            segment_index, tm.dependent_map.shape[-1],
        )
        for k in hd5[tm.path_prefix]:
            if k in tm.channel_map:
                tensor[:, tm.channel_map[k]] = np.array(hd5[tm.path_prefix][k], dtype=np.float32)[
                                                        offset_begin:offset_begin+tm.shape[0]
                ]
        if population_normalize is None:
            tm.normalization = {'zero_mean_std1': 1.0}
        else:
            tensor /= population_normalize
        return tensor
    return ecg_rest_section_to_segment


ecg_bike_hrr = TensorMap(
    'hrr', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': 30.55, 'std': 12.81},
    tensor_from_file=_first_date_hrr,
)
ecg_bike_healthy_max_hr = TensorMap(
    'max_hr', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'],
    normalization={'mean': 113.7, 'std': 13.3}, shape=(1,),
    tensor_from_file=_healthy_bike,
)
ecg_bike_healthy_hrr = TensorMap(
    'hrr', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': 30.47, 'std': 11.76},
    tensor_from_file=_healthy_hrr,
)
ecg_bike_healthy_resting = TensorMap(
    'resting_hr', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': 70.0, 'std': 11.62},
    tensor_from_file=_healthy_bike,
)
ecg_bike_med_pretest_hr = TensorMap(
    'trend_heartrate', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': 70., 'std': 11.},
    tensor_from_file=_median_pretest,
)
ecg_bike_med_pretest_stamp = TensorMap(
    'trend_stamplitude', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': .03, 'std': .03},
    tensor_from_file=_median_pretest,
)
ecg_bike_med_pretest_jpoint = TensorMap(
    'trend_jpointamplitude', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': .032, 'std': .46},
    tensor_from_file=_median_pretest,
)
ecg_bike_med_pretest_stamp20 = TensorMap(
    'trend_stamplitude20ms', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': .03, 'std': .03},
    tensor_from_file=_median_pretest,
)
ecg_bike_recovery = TensorMap(
    'full', shape=(30000, 1), path_prefix='ecg_bike', validator=no_nans,
    tensor_from_file=_first_date_bike_recovery,
)
ecg_bike_pretest = TensorMap(
    'full', shape=(500 * 15 - 4, 3), path_prefix='ecg_bike', validator=no_nans,
    normalization={
        'mean': np.array(
        [7, -7, 3.5],
        )[np.newaxis], 'std': np.array([31, 30, 16])[np.newaxis],
    },
    tensor_from_file=_first_date_bike_pretest,
)
ecg_bike_pretest_5k = TensorMap(
    'full', shape=(5000, 3), path_prefix='ecg_bike', validator=no_nans,
    normalization={
        'mean': np.array(
        [7, -7, 3.5],
        )[np.newaxis], 'std': np.array([31, 30, 16])[np.newaxis],
    },
    tensor_from_file=_first_date_bike_pretest,
)
ecg_bike_new_hrr = TensorMap(
    'hrr', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': 31, 'std': 12},
    tensor_from_file=_new_hrr,
)
ecg_bike_hrr_sentinel = TensorMap(
    'hrr', path_prefix='ecg_bike', metrics=['mae'], shape=(1,),
    normalization={'mean': 31, 'std': 12}, sentinel=_HRR_SENTINEL,
    tensor_from_file=_sentinel_hrr,
)
ecg_bike_hrr_student = TensorMap(
    'hrr', path_prefix='ecg_bike', metrics=['mae'], shape=(1,),
    normalization={'mean': 31, 'std': 12}, sentinel=_HRR_SENTINEL,
    tensor_from_file=build_tensor_from_file(
        'inference.tsv', 'ecg_bike_hrr-sentinel_prediction',
    ),
)
ecg_bike_hr_achieved = TensorMap(
    'hr_achieved', path_prefix='ecg_bike', loss='logcosh', metrics=['mae'], shape=(1,),
    normalization={'mean': .68, 'std': .1},
    tensor_from_file=_hr_achieved,
)

ecg_rest_raw = TensorMap(
    'ecg_rest_raw', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=Standardize(mean=0, std=2000),
)

ecg_rest_raw_mv = TensorMap(
    'ecg_rest_raw_mv', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=Standardize(mean=0, std=1000),
)
ecg_rest_raw_mv_warp = TensorMap(
    'ecg_rest_raw_mv_warp', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=Standardize(mean=0, std=1000), augmentations=[_warp_ecg],
)

ecg_rest_raw_100 = TensorMap(
    'ecg_rest_raw_100', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=Standardize(mean=0, std=100),
)
ecg_rest_raw_8s = TensorMap(
    'ecg_rest_raw_8s', Interpretation.CONTINUOUS, shape=(4096, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=Standardize(mean=0, std=10),
)

ecg_rest_raw_10 = TensorMap(
    'ecg_rest_raw_10', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=Standardize(mean=0, std=10),
)


ecg_rest = TensorMap(
    'strip', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(),
)
ecg_rest_8s = TensorMap(
    'ecg_rest_8s', Interpretation.CONTINUOUS, shape=(4096, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(),
)
ecg_rest_no_poor = TensorMap(
    'strip', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(skip_poor=True),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(),
)

ecg_rest_mgb = TensorMap(
    'ecg_5000_std', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_AMP_LEADS_UKB, normalization=ZeroMeanStd1(),
)

ecg_rest_mgb_mv = TensorMap(
    'ecg_5000_mv', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_AMP_LEADS_UKB, normalization=Standardize(mean=0, std=150),
)
mgb_ecg_rest_mv = TensorMap(
    'mgb_ecg_5000_mv', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_AMP_LEADS_UKB, normalization=Standardize(mean=0, std=150),
)
ecg_rest_mgb_2500 = TensorMap(
    'ecg_2500_std', Interpretation.CONTINUOUS, shape=(2500, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(downsample_steps=2),
    channel_map=ECG_REST_AMP_LEADS_UKB, normalization=ZeroMeanStd1(),
)
mgb_ecg_rest_2500 = TensorMap(
    'mgb_ecg_2500_std', Interpretation.CONTINUOUS, shape=(2500, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(downsample_steps=2),
    channel_map=ECG_REST_AMP_LEADS_UKB, normalization=ZeroMeanStd1(),
)
partners_ecg_2500 = TensorMap(
    'partners_ecg_2500_newest', Interpretation.CONTINUOUS, shape=(2500, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_AMP_LEADS_UKB, normalization=ZeroMeanStd1(),
)

ecg_rest_mgb_strip_I = TensorMap(
    'ecg_strip_I', Interpretation.CONTINUOUS, shape=(5000, 1), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map={'strip_I': 0}, normalization=ZeroMeanStd1(),
)
ecg_rest_mgb_strip_II = TensorMap(
    'ecg_strip_II', Interpretation.CONTINUOUS, shape=(5000, 1), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map={'strip_II': 0}, normalization=ZeroMeanStd1(),
)

rest_strip_I_random = TensorMap(
    'ecg_strip_I', Interpretation.CONTINUOUS, shape=(4096, 1), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(random_offset=904),
    channel_map={'strip_I': 0}, normalization=ZeroMeanStd1(),
)
rest_strip_II_random = TensorMap(
    'ecg_strip_II', Interpretation.CONTINUOUS, shape=(4096, 1), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(random_offset=904),
    channel_map={'strip_II': 0}, normalization=ZeroMeanStd1(),
)
rest_strip_I_random_1s = TensorMap(
    'rest_strip_I_random_1s', Interpretation.CONTINUOUS, shape=(500, 1), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(random_offset=4500),
    channel_map={'strip_I': 0}, normalization=ZeroMeanStd1(),
)
rest_strip_I_random_2s = TensorMap(
    'rest_strip_I_random_2s', Interpretation.CONTINUOUS, shape=(1000, 1), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(random_offset=4000),
    channel_map={'strip_I': 0}, normalization=ZeroMeanStd1(),
)
rest_strip_II_random_2s = TensorMap(
    'rest_strip_II_random_2s', Interpretation.CONTINUOUS, shape=(1000, 1), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(random_offset=4000),
    channel_map={'strip_II': 0}, normalization=ZeroMeanStd1(),
)
ecg_rest_mgb_2500_test = TensorMap(
    'ecg_2500_std', Interpretation.CONTINUOUS, shape=(2500, 12), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(downsample_steps=2), channel_map=ECG_REST_AMP_LEADS_UKB, normalization=ZeroMeanStd1(),
)

ecg_rest_mgb_2500_raw = TensorMap(
    'partners_ecg_2500_newest', Interpretation.CONTINUOUS, shape=(2500, 12), path_prefix='ukb_ecg_rest',
    tensor_from_file=_make_ecg_rest(downsample_steps=2), channel_map=ECG_REST_MGB_LEADS, normalization=Standardize(mean=0, std=2000),
)
ecg_rest_mse10x = TensorMap(
    'strip', Interpretation.CONTINUOUS, shape=(5000, 12), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(), loss=mse_10x,
)
ecg_rest_2500_ukb = TensorMap(
    'ecg_rest_2500', Interpretation.CONTINUOUS, shape=(2500, 12), path_prefix='ukb_ecg_rest', channel_map=ECG_REST_LEADS,
    tensor_from_file=_make_ecg_rest(downsample_steps=2), normalization=ZeroMeanStd1(),
)

ecg_rest_stft = TensorMap(
    'ecg_rest_stft', Interpretation.CONTINUOUS, shape=(52, 72, 12), path_prefix='ukb_ecg_rest', channel_map=ECG_REST_LEADS,
    tensor_from_file=_make_ecg_rest(short_time_nperseg=103, short_time_noverlap=32), normalization=ZeroMeanStd1(),
)
ecg_rest_stft_512 = TensorMap(
    'ecg_rest_stft_512', shape=(257, 314, 12), path_prefix='ukb_ecg_rest', channel_map=ECG_REST_LEADS,
    tensor_from_file=_make_ecg_rest(short_time_nperseg=512, short_time_noverlap=496), normalization=ZeroMeanStd1(),
)

ecg_rest_stack = TensorMap(
    'strip', Interpretation.CONTINUOUS, shape=(600, 12, 8), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(),
)

ecg_rest_median_raw = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh', activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=Standardize(mean=0, std=2000),
)
ecg_rest_median_lead_I = TensorMap(
    'ecg_rest_median_lead_I', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest',
    shape=(600, 1), loss='logcosh', activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae'], channel_map={'median_I': 0}, normalization=ZeroMeanStd1(),
)
ecg_rest_median_lead_II = TensorMap(
    'ecg_rest_median_lead_II', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest',
    shape=(600, 1), loss='logcosh', activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae'], channel_map={'median_II': 0}, normalization=ZeroMeanStd1(),
)
ecg_rest_median_mv = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh', activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=Standardize(mean=0, std=1000),
)
ecg_rest_median_mv_warp = TensorMap(
    'median_warp', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh', activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=Standardize(mean=0, std=1000), augmentations=[_warp_ecg],
)

ecg_rest_median_raw_10 = TensorMap(
    'ecg_rest_median_raw_10', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh',
    activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=Standardize(mean=0, std=10),
)

ecg_rest_median_576 = TensorMap(
    'ecg_rest_median_576', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(576, 12), loss='logcosh',
    activation='linear', tensor_from_file=_make_ecg_rest(), channel_map=ECG_REST_MEDIAN_LEADS,
    normalization=Standardize(mean=0, std=10),
)

ecg_rest_median_raw_10_no_poor = TensorMap(
    'ecg_rest_median_raw_10', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh', activation='linear',
    tensor_from_file=_make_ecg_rest(skip_poor=True), metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=Standardize(mean=0, std=10),
)
ecg_rest_median_raw_10_random_norm = TensorMap(
    'ecg_rest_median_raw_10', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh', activation='linear',
    tensor_from_file=_make_ecg_rest(skip_poor=True),
    metrics=['mse', 'mae', 'logcosh'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=RandomStandardize(mean=0, std=10),
)
ecg_rest_median = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), loss='logcosh', activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae', 'logcosh'], channel_map=ECG_REST_MEDIAN_LEADS, normalization=ZeroMeanStd1(),
)

ecg_rest_median_raw_10_prediction = TensorMap(
    'ecg_rest_median_raw_10', Interpretation.CONTINUOUS, shape=(600, 12), loss='logcosh', activation='linear', normalization=ZeroMeanStd1(),
    tensor_from_file=named_tensor_from_hd5('ecg_rest_median_raw_10_prediction'), metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS,
)


def stretch_ecg(x, n=0):
    """
    stretches input ECG to n bpm
    """
    out = biosppy.signals.ecg.ecg(x.copy(), sampling_rate=500, show=False)
    hr = out[-1].mean()
    t = np.arange(len(x))
    if n == 0:
        tp = np.arange(len(x))
    else:
        tp = np.arange(len(x)) * n / hr
    stretched = np.interp(tp, t, x)
    out2 = biosppy.signals.ecg.ecg(stretched, show=False)
    return stretched, out2[2]


def make_biosspy_median(example, channel_map, median_size = 600, bpm = 0):
    medians = np.zeros((median_size, len(channel_map)))
    for lead in channel_map:
        waves = []

        stretched, peaks = stretch_ecg(example[:,channel_map[lead]], bpm)

        for j, p0 in enumerate(peaks[:-2]):
            p11 = peaks[j+1]
            middle = (p0+p11)//2
            waves.append(stretched[middle:middle+median_size])
        waves = np.array(waves)
        medians[:, channel_map[lead]] = np.median(waves, axis=0)

    return medians


def ecg_median_biosppy(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    tensor = np.zeros(tm.shape, dtype=np.float32)
    for lead in tm.channel_map:
        tensor[:, tm.channel_map[lead]] = hd5[f'{tm.path_prefix}{lead}']
    tensor = np.nan_to_num(tensor)
    return tensor


ecg_biosppy_median = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_', shape=(600, 12), tensor_from_file=ecg_median_biosppy,
    channel_map=ECG_REST_LEADS,
)

ecg_biosppy_median_60bpm = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_60bpm_', shape=(600, 12), tensor_from_file=ecg_median_biosppy,
    channel_map=ECG_REST_LEADS,
)

ecg_biosppy_median_60bpm_lead_I = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_60bpm_', shape=(600, 12), tensor_from_file=ecg_median_biosppy,
    channel_map={'strip_I': 0},
)

ecg_biosppy_median_mgb = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_', shape=(600, 12), tensor_from_file=ecg_median_biosppy,
    channel_map=ECG_REST_AMP_LEADS,
)

ecg_biosppy_median_60bpm_mgb = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_60bpm_', shape=(600, 12), tensor_from_file=ecg_median_biosppy,
    channel_map=ECG_REST_AMP_LEADS,
)
ecg_biosppy_median_60bpm_mgb_lead_I = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='median_60bpm_', shape=(600, 12),
    tensor_from_file=ecg_median_biosppy,
    channel_map={'I': 0},
)


def ecg_median_biosppy_on_the_fly(ecg_10s_shape=(5000, 12), bpm=0, instance=2):
    def _ecg_median_tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
        ecg_10s = np.zeros(ecg_10s_shape, dtype=np.float32)
        for k in hd5[tm.path_prefix]:
            if k in tm.channel_map:
                data = tm.hd5_first_dataset_in_group(
                    hd5, f'{tm.path_prefix}/{k}/instance_{instance}',
                )
                ecg_10s[:, tm.channel_map[k]] = pad_or_crop_array_to_shape((ecg_10s_shape[0],), data)
        return make_biosspy_median(ecg_10s, tm.channel_map, median_size=tm.shape[0], bpm=bpm)
    return _ecg_median_tensor_from_file


ecg_biosppy_median_otf = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), tensor_from_file=ecg_median_biosppy_on_the_fly(),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(),
)
ecg_biosppy_median_60bpm_otf = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12), tensor_from_file=ecg_median_biosppy_on_the_fly(bpm=60),
    channel_map=ECG_REST_LEADS, normalization=ZeroMeanStd1(),
)


def ecg_prediction_lead_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    ecg = np.array(hd5['ecg_rest_median_raw_10_prediction'], dtype=np.float32)
    tensor = np.zeros(tm.shape, dtype=np.float32)
    tensor[:, 0] = ecg[:, 0]
    return tensor


ecg_rest_median_raw_10_prediction_lead_I = TensorMap(
    'ecg_rest_median_raw_10_lead_I', Interpretation.CONTINUOUS, shape=(600, 1), loss='logcosh', activation='linear',
    tensor_from_file=ecg_prediction_lead_from_hd5, metrics=['mse', 'mae'], channel_map={'median_I': 0},
)

ecg_rest_median_raw_10_truth = TensorMap(
    'ecg_rest_median_raw_10_truth', Interpretation.CONTINUOUS, shape=(600, 12), loss='logcosh', activation='linear',
    tensor_from_file=tensor_from_hd5, metrics=['mse', 'mae'], channel_map=ECG_REST_MEDIAN_LEADS,
)

ecg_rest_median_stack = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 12, 1), activation='linear', tensor_from_file=_make_ecg_rest(),
    metrics=['mse', 'mae', 'logcosh'], loss='logcosh', loss_weight=1.0,
    channel_map=ECG_REST_MEDIAN_LEADS, normalization=ZeroMeanStd1(),
)

ecg_median_1lead = TensorMap(
    'median', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(600, 1), loss='logcosh', loss_weight=10.0, tensor_from_file=_make_ecg_rest(),
    activation='linear', metrics=['mse', 'mae', 'logcosh'], channel_map={'lead': 0}, normalization=ZeroMeanStd1(),
)

ecg_rest_1lead = TensorMap(
    'strip', Interpretation.CONTINUOUS, shape=(600, 8), path_prefix='ukb_ecg_rest', channel_map={'lead': 0}, tensor_from_file=_make_ecg_rest(),
    dependent_map=ecg_median_1lead, normalization=ZeroMeanStd1(),
)

ecg_median_1lead_categorical = TensorMap(
    'median',  Interpretation.CATEGORICAL, shape=(600, 32), activation='softmax', tensor_from_file=_make_ecg_rest(),
    channel_map=_get_lead_cm(32)[0],  normalization=ZeroMeanStd1(),
    loss=weighted_crossentropy(
        np.array(_get_lead_cm(32)[1]), 'ecg_median_categorical',
    ),
)

ecg_rest_1lead_categorical = TensorMap(
    'strip', shape=(600, 8), path_prefix='ukb_ecg_rest', tensor_from_file=_make_ecg_rest(),
    normalization=ZeroMeanStd1(),
    channel_map={
        'window0': 0, 'window1': 1, 'window2': 2, 'window3': 3,
        'window4': 4, 'window5': 5, 'window6': 6, 'window7': 7,
    },
    dependent_map=ecg_median_1lead_categorical,
)

ecg_rhythm = TensorMap(
    'ecg_rhythm', Interpretation.CATEGORICAL, tensor_from_file=_make_rhythm_tensor(),
    loss=weighted_crossentropy([1.0, 2.0, 3.0, 3.0, 20.0, 20.0], 'ecg_rhythm'),
    channel_map={
        'Normal_sinus_rhythm': 0, 'Sinus_bradycardia': 1, 'Marked_sinus_bradycardia': 2,
        'Other_sinus_rhythm': 3, 'Atrial_fibrillation': 4, 'Other_rhythm': 5,
    },
)
ecg_rhythm_poor = TensorMap(
    'ecg_rhythm', Interpretation.CATEGORICAL, tensor_from_file=_make_rhythm_tensor(False),
    loss=weighted_crossentropy(
        [1.0, 2.0, 3.0, 3.0, 20.0, 20.0], 'ecg_rhythm_poor',
    ),
    channel_map={
        'Normal_sinus_rhythm': 0, 'Sinus_bradycardia': 1, 'Marked_sinus_bradycardia': 2,
        'Other_sinus_rhythm': 3, 'Atrial_fibrillation': 4, 'Other_rhythm': 5,
    },
)

ecg_rest_age = TensorMap(
    'ecg_rest_age', Interpretation.CONTINUOUS, tensor_from_file=age_in_years_tensor('ecg_rest_date'), loss='logcosh',
    channel_map={'ecg_rest_age': 0}, validator=make_range_validator(0, 110), normalization={'mean': 65, 'std': 7.7},
)

acute_mi = TensorMap(
    'acute_mi', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text, channel_map={'no_acute_mi': 0, 'ACUTE MI': 1},
    loss=weighted_crossentropy([0.1, 10.0], 'acute_mi'),
)

anterior_blocks = TensorMap(
    'anterior_blocks', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_anterior_blocks': 0, 'Left anterior fascicular block': 1,
        'Left posterior fascicular block': 2,
    },
    loss=weighted_crossentropy([0.1, 10.0, 10.0], 'anterior_blocks'),
)

av_block = TensorMap(
    'av_block', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text, channel_map={'no_av_block': 0, 'st degree AV block': 1},
    loss=weighted_crossentropy([0.1, 10.0], 'av_block'),
)

incomplete_right_bundle_branch_block = TensorMap(
    'incomplete_right_bundle_branch_block', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_incomplete_right_bundle_branch_block': 0,
        'Incomplete right bundle branch block': 1,
    },
    loss=weighted_crossentropy(
        [0.1, 10.0], 'incomplete_right_bundle_branch_block',
    ),
)

infarcts = TensorMap(
    'infarcts', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_infarcts': 0, 'Anterior infarct': 1, 'Anteroseptal infarct': 2,
        'Inferior infarct': 3, 'Lateral infarct': 4, 'Septal infarct': 5,
    },
    loss=weighted_crossentropy([0.1, 4.0, 6.0, 7.0, 6.0, 4.0], 'infarcts'),
)

left_atrial_enlargement = TensorMap(
    'left_atrial_enlargement', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_left_atrial_enlargement': 0,
        'Left atrial enlargement': 1,
    },
    loss=weighted_crossentropy([0.1, 10.0], 'left_atrial_enlargement'),
)

left_ventricular_hypertrophy = TensorMap(
    'left_ventricular_hypertrophy', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_left_ventricular_hypertrophy': 0,
        'Left ventricular hypertrophy': 1,
    },
    loss=weighted_crossentropy([0.1, 10.0], 'left_ventricular_hypertrophy'),
)

lvh_fine = TensorMap(
    'lvh_fine', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text, loss=weighted_crossentropy([0.5, 12.0, 16.0, 30.0, 36.0], 'lvh_fine'),
    channel_map={
        'no_lvh_fine': 0, 'Minimal voltage criteria for LVH may be normal variant': 1,
        'Moderate voltage criteria for LVH may be normal variant': 2, 'Voltage criteria for left ventricular hypertrophy': 3,
        'Left ventricular hypertrophy': 4,
    },
)



premature_atrial_complexes = TensorMap(
    'premature_atrial_complexes', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_premature_atrial_complexes': 0,
        'premature atrial complexes': 1,
    },
    loss=weighted_crossentropy([0.1, 10.0], 'premature_atrial_complexes'),
)

premature_supraventricular_complexes = TensorMap(
    'premature_supraventricular_complexes', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_premature_supraventricular_complexes': 0,
        'premature supraventricular complexes': 1,
    },
    loss=weighted_crossentropy(
        [0.1, 10.0], 'premature_supraventricular_complexes',
    ),
)

premature_ventricular_complexes = TensorMap(
    'premature_ventricular_complexes', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={
        'no_premature_ventricular_complexes': 0,
        'premature ventricular complexes': 1,
    },
    loss=weighted_crossentropy([0.1, 10.0], 'premature_ventricular_complexes'),
)

prolonged_qt = TensorMap(
    'prolonged_qt', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text, channel_map={'no_prolonged_qt': 0, 'Prolonged QT': 1},
    loss=weighted_crossentropy([0.1, 10.0], 'prolonged_qt'),
)


ecg_rest_ramplitude_raw = TensorMap(
    'ramplitude', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(12,), tensor_from_file=_make_ukb_ecg_rest(1.0),
    loss='logcosh', metrics=['mse', 'mape', 'mae'], loss_weight=1.0,
)

ecg_rest_samplitude_raw = TensorMap(
    'samplitude', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(12,), tensor_from_file=_make_ukb_ecg_rest(1.0),
    loss='logcosh', metrics=['mse', 'mape', 'mae'], loss_weight=1.0,
)

ecg_rest_ramplitude = TensorMap(
    'ramplitude', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(12,), tensor_from_file=_make_ukb_ecg_rest(),
    loss='logcosh', metrics=['mse', 'mape', 'mae'], loss_weight=1.0,
)

ecg_rest_samplitude = TensorMap(
    'samplitude', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', shape=(12,), tensor_from_file=_make_ukb_ecg_rest(),
    loss='logcosh', metrics=['mse', 'mape', 'mae'], loss_weight=1.0,
)


ecg_rest_lvh_avl = TensorMap(
    'avl_lvh', Interpretation.CATEGORICAL, path_prefix='ukb_ecg_rest', tensor_from_file=_make_ukb_ecg_rest_lvh(),
    channel_map={'no_avl_lvh': 0, 'aVL LVH': 1},
    loss=weighted_crossentropy([0.006, 1.0], 'avl_lvh'),
)

ecg_rest_lvh_sokolow_lyon = TensorMap(
    'sokolow_lyon_lvh', Interpretation.CATEGORICAL, path_prefix='ukb_ecg_rest', tensor_from_file=_make_ukb_ecg_rest_lvh(),
    channel_map={'no_sokolow_lyon_lvh': 0, 'Sokolow Lyon LVH': 1},
    loss=weighted_crossentropy([0.005, 1.0], 'sokolov_lyon_lvh'),
)

ecg_rest_lvh_cornell = TensorMap(
    'cornell_lvh', Interpretation.CATEGORICAL, path_prefix='ukb_ecg_rest', tensor_from_file=_make_ukb_ecg_rest_lvh(),
    channel_map={'no_cornell_lvh': 0, 'Cornell LVH': 1},
    loss=weighted_crossentropy([0.003, 1.0], 'cornell_lvh'),
)


ecg_segmented = TensorMap(
    'ecg_segmented', Interpretation.CATEGORICAL, shape=(1224, len(ECG_SEGMENTED_CHANNEL_MAP)), path_prefix='ecg_rest',
    cacheable=False, channel_map=ECG_SEGMENTED_CHANNEL_MAP,
)
ecg_section_to_segment = TensorMap(
    'ecg_section_to_segment', shape=(1224, 12), path_prefix='ecg_rest', dependent_map=ecg_segmented,
    channel_map=ECG_REST_LEADS, tensor_from_file=_ecg_rest_to_segment(),
)
ecg_section_to_segment_warp = TensorMap(
    'ecg_section_to_segment', shape=(1224, 12), path_prefix='ecg_rest', dependent_map=ecg_segmented,
    cacheable=False, channel_map=ECG_REST_LEADS, tensor_from_file=_ecg_rest_to_segment(),
    augmentations=[_warp_ecg],
)

ecg_segmented_second = TensorMap(
    'ecg_segmented', Interpretation.CATEGORICAL, shape=(496, len(ECG_SEGMENTED_CHANNEL_MAP)), path_prefix='ecg_rest',
    cacheable=False, channel_map=ECG_SEGMENTED_CHANNEL_MAP,
)
ecg_second_to_segment = TensorMap(
    'ecg_second_to_segment', shape=(496, 12), path_prefix='ecg_rest', dependent_map=ecg_segmented_second,
    cacheable=False, channel_map=ECG_REST_LEADS, tensor_from_file=_ecg_rest_to_segment(random_offset_seconds=1.5),
)
ecg_second_to_segment_warp = TensorMap(
    'ecg_second_to_segment', shape=(496, 12), path_prefix='ecg_rest', dependent_map=ecg_segmented_second,
    cacheable=False, channel_map=ECG_REST_LEADS, tensor_from_file=_ecg_rest_to_segment(random_offset_seconds=1.5),
    augmentations=[_warp_ecg],
)

poor_data_quality = TensorMap(
    'poor_data_quality', Interpretation.CATEGORICAL, tensor_from_file=label_from_ecg_interpretation_text, channel_map={'no_poor_data_quality': 0, 'Poor data quality': 1},
    loss=weighted_crossentropy([0.1, 3.0], 'poor_data_quality'),
)

####
ecg_semi_coarse = TensorMap(
    'ecg_semi_coarse', Interpretation.CATEGORICAL, loss=weighted_crossentropy([1.0, 1.0, 2.0, 4.0, 16.0, 20.0], 'ecg_semi_coarse'),
    channel_map={'Normal_sinus_rhythm': 0, 'Sinus_bradycardia': 1, 'Marked_sinus_bradycardia': 2, 'Other_sinus_rhythm': 3, 'Atrial_fibrillation': 4, 'Other_rhythm': 5},
)


ecg_semi_coarse_with_poor = TensorMap(
    'ecg_semi_coarse_with_poor', Interpretation.CATEGORICAL, loss=weighted_crossentropy([1.0, 2.0, 3.0, 3.0, 20.0, 20.0], 'ecg_semi_coarse_with_poor'),
    channel_map={'Normal_sinus_rhythm': 0, 'Sinus_bradycardia': 1, 'Marked_sinus_bradycardia': 2, 'Other_sinus_rhythm': 3, 'Atrial_fibrillation': 4, 'Other_rhythm': 5},
)

normal = TensorMap(
    'normal', Interpretation.CATEGORICAL, loss=weighted_crossentropy([2.0, 3.0, 3.0, 3.0], 'ecg_normal'),
    tensor_from_file=label_from_ecg_interpretation_text,
    channel_map={'Normal ECG': 0, 'Abnormal ECG': 1, 'Borderline ECG': 2, 'Otherwise normal ECG': 3},
)
infarct = TensorMap(
    'infarct', Interpretation.CATEGORICAL, channel_map={'no_infarct': 0, 'infarct': 1},
    tensor_from_file=label_from_ecg_interpretation_text,
    loss=weighted_crossentropy([1.0, 8.0], 'ecg_infarct'),
)
ecg_poor_data = TensorMap(
    'ecg_poor_data', Interpretation.CATEGORICAL, channel_map={'no_poor_data_quality': 0, 'poor_data_quality': 1},
    tensor_from_file=label_from_ecg_interpretation_text,
    loss=weighted_crossentropy([1.0, 8.0], 'ecg_poor_data'),
)
block = TensorMap(
    'block', Interpretation.CATEGORICAL, channel_map={'no_block': 0, 'block': 1},
    tensor_from_file=label_from_ecg_interpretation_text,
    loss=weighted_crossentropy([1.0, 8.0], 'ecg_block'),
)

ecg_rest_next_char = TensorMap('ecg_rest_next_char', Interpretation.LANGUAGE, shape=(len(ECG_CHAR_2_IDX),), channel_map=ECG_CHAR_2_IDX, activation='softmax', loss='categorical_crossentropy', loss_weight=2.0)
ecg_rest_text = TensorMap('ecg_rest_text', Interpretation.LANGUAGE, shape=(100, len(ECG_CHAR_2_IDX)), path_prefix='ukb_ecg_rest', channel_map={'context': 0, 'alphabet': 1}, dependent_map=ecg_rest_next_char)

p_axis = TensorMap(
    'PAxis', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PAxis': 0}, loss='logcosh', validator=make_range_validator(-50, 130),
    normalization={'mean': 48.7, 'std': 23.1},
)
p_duration = TensorMap(
    'PDuration', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PDuration': 0}, loss='logcosh', validator=make_range_validator(30, 140),
    normalization={'mean': 96.1, 'std': 18.85},
)
p_offset = TensorMap(
    'POffset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'POffset': 0}, loss='logcosh', validator=make_range_validator(200, 500),
    normalization={'mean': 369.1, 'std': 28.42},
)
p_onset = TensorMap(
    'POnset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'POnset': 0}, loss='logcosh', validator=make_range_validator(120, 400),
    normalization={'mean': 275.1, 'std': 26.420},
)
pp_interval = TensorMap(
    'PPInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PPInterval': 0}, loss='logcosh', validator=make_range_validator(300, 1800),
    normalization={'mean': 1036.1, 'std': 185.0},
)
pq_interval = TensorMap(
    'PQInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PQInterval': 0}, loss='logcosh', validator=make_range_validator(70, 400),
    normalization={'mean': 165.9, 'std': 26.3},
)
q_offset = TensorMap(
    'QOffset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QOffset': 0}, loss='logcosh', validator=make_range_validator(300, 600),
    normalization={'mean': 525.1, 'std': 13.52},
)
q_onset = TensorMap(
    'QOnset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QOnset': 0}, loss='logcosh', validator=make_range_validator(370, 600),
    normalization={'mean': 435.1, 'std': 11.420},
)
qrs_complexes = TensorMap(
    'QRSComplexes', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QRSComplexes': 0}, loss='logcosh', validator=make_range_validator(0, 60),
    normalization={'mean': 8.0, 'std': 20.0},
)
qrs_duration = TensorMap(
    'QRSDuration', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QRSDuration': 0}, loss='logcosh', validator=make_range_validator(45, 175),
    normalization={'mean': 89.53, 'std': 12.21},
)
qrs_num = TensorMap(
    'QRSNum', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QRSNum': 0}, loss='logcosh', validator=make_range_validator(2, 30),
    normalization={'mean': 9.61, 'std': 1.64},
)
qt_interval = TensorMap(
    'QTInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTInterval': 0}, loss='logcosh', validator=make_range_validator(300, 600),
    normalization=Standardize(mean=426.1, std=32.24),
)
qt_interval_quintiles = TensorMap(
    'QTInterval', Interpretation.DISCRETIZED, path_prefix='ukb_ecg_rest',
    channel_map={'QTInterval': 0}, normalization={'mean': 426.1, 'std': 32.24},
    discretization_bounds=[-0.842, -0.253, 0.253, 0.842],
)
qtc_interval = TensorMap(
    'QTCInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTCInterval': 0}, loss='logcosh', validator=make_range_validator(300, 600),
    normalization={'mean': 419.1, 'std': 20.7},
)
r_axis = TensorMap(
    'RAxis', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'RAxis': 0}, loss='logcosh', validator=make_range_validator(-100, 200),
    normalization={'mean': 25.7, 'std': 36.6},
)
rr_interval = TensorMap(
    'RRInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'RRInterval': 0}, loss='logcosh',
    validator=make_range_validator(400, 2000), normalization={'mean': 1040.61, 'std': 175.5}, metrics=[],
)
ventricular_rate = TensorMap(
    'VentricularRate', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'VentricularRate': 0}, validator=make_range_validator(30, 150),
    loss='logcosh', normalization={'mean': 59.3, 'std': 10.6},
)
t_offset = TensorMap(
    'TOffset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'TOffset': 0}, loss='logcosh', validator=make_range_validator(700, 1000),
    normalization={'mean': 860.7, 'std': 32.52},
)
t_axis = TensorMap(
    'TAxis', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'TAxis': 0}, loss='logcosh', validator=make_range_validator(-100, 200),
    normalization={'mean': 40.8, 'std': 32.6},
)

af_prs = TensorMap('AF_PRS_LDscore', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'AF_PRS_LDscore': 0}, normalization={'mean': -1.0, 'std': 0.4})
charge = TensorMap(
    'charge', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'charge': 0}, normalization={'mean': 12.0, 'std': 2.0},
    validator=make_range_validator(0, 20),
)

qtc_intervalp = TensorMap(
    'QTCInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTCInterval': 0}, loss='logcosh', validator=make_range_validator(100, 900),
    parents=[qt_interval, rr_interval], normalization={'mean': 419.1, 'std': 20.7},
)
qrs_durationpp = TensorMap(
    'QRSDuration', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QRSDuration': 0}, loss='logcosh', validator=make_range_validator(45, 175),
    normalization={'mean': 89.53, 'std': 12.21},
    parents=[qtc_intervalp],
)

p_axis_sentinel = TensorMap(
    'PAxis', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PAxis': 0}, sentinel=0, metrics=['logcosh'],
    normalization={'mean': 48.7, 'std': 23.1},
)
p_duration_sentinel = TensorMap(
    'PDuration', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PDuration': 0}, sentinel=0, metrics=['logcosh'],
    normalization={'mean': 96.1, 'std': 18.85},
)
p_offset_sentinel = TensorMap(
    'POffset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'POffset': 0}, sentinel=0, metrics=['logcosh'],
    normalization={'mean': 369.1, 'std': 28.42},
)
p_onset_sentinel = TensorMap(
    'POnset', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'POnset': 0}, sentinel=0, metrics=['logcosh'],
    normalization={'mean': 275.1, 'std': 26.420},
)
pp_interval_sentinel = TensorMap(
    'PPInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PPInterval': 0}, sentinel=0, metrics=['logcosh'],
    normalization={'mean': 1036.1, 'std': 185.0},
)
pq_interval_sentinel = TensorMap(
    'PQInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'PQInterval': 0}, sentinel=0, metrics=['logcosh'],
    normalization={'mean': 165.9, 'std': 26.3},
)
qrs_duration_sentinel = TensorMap(
    'QRSDuration', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QRSDuration': 0}, sentinel=0,
    normalization={'mean': 89.53, 'std': 12.21},
)
qt_interval_sentinel = TensorMap(
    'QTInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTInterval': 0}, sentinel=0,
    normalization={'mean': 426.1, 'std': 32.24},
)
qtc_interval_sentinel = TensorMap(
    'QTCInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTCInterval': 0}, sentinel=0,
    normalization={'mean': 419.1, 'std': 20.7},
)
qtc_intervalp_sentinel = TensorMap(
    'QTCInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTCInterval': 0}, sentinel=0,
    normalization={'mean': 419.1, 'std': 20.7},
    parents=[qt_interval, rr_interval],
)
qtc_intervalp_sentinel = TensorMap(
    'QTCInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'QTCInterval': 0}, sentinel=0,
    normalization={'mean': 419.1, 'std': 20.7},
    parents=[qt_interval, rr_interval],
)
r_axis_sentinel = TensorMap('RAxis', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'RAxis': 0}, sentinel=0, normalization={'mean': 25.7, 'std': 36.6})
rr_interval_sentinel = TensorMap(
    'RRInterval', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'RRInterval': 0}, sentinel=0,
    normalization={'mean': 1040.61, 'std': 175.5},
)
t_axis_sentinel = TensorMap('TAxis', Interpretation.CONTINUOUS, path_prefix='ukb_ecg_rest', channel_map={'TAxis': 0}, sentinel=0, normalization={'mean': 40.8, 'std': 32.6})


bb_baseline = TensorMap(
    'bb_baseline', Interpretation.CATEGORICAL, channel_map={'no_bb_baseline': 0, 'bb_baseline': 1},
    loss=weighted_crossentropy([0.0453, 0.9547], 'bb_baseline'),
)
ccb_baseline = TensorMap(
    'ccb_baseline', Interpretation.CATEGORICAL, channel_map={'no_ccb_baseline': 0, 'ccb_baseline': 1},
    loss=weighted_crossentropy([0.0044, 0.9956], 'ccb_baseline'),
)
class1_baseline = TensorMap(
    'class1_baseline', Interpretation.CATEGORICAL, channel_map={'no_class1_baseline': 0, 'class1_baseline': 1},
    loss=weighted_crossentropy([0.0023, 0.9977], 'class1_baseline'),
)
class3_baseline = TensorMap(
    'class3_baseline', Interpretation.CATEGORICAL, channel_map={'no_class3_baseline': 0, 'class3_baseline': 1},
    loss=weighted_crossentropy([0.0011, 0.9989], 'class3_baseline'),
)
qtc_drug_def_baseline = TensorMap(
    'qtc_drug_def_baseline', Interpretation.CATEGORICAL,
    channel_map={'no_qtc_drug_def_baseline': 0, 'qtc_drug_def_baseline': 1},
    loss=weighted_crossentropy([0.0210, 0.9790], 'qtc_drug_def_baseline'),
)
qtc_drug_poss_baseline = TensorMap(
    'qtc_drug_poss_baseline', Interpretation.CATEGORICAL,
    channel_map={'no_qtc_drug_poss_baseline': 0, 'qtc_drug_poss_baseline': 1},
    loss=weighted_crossentropy([0.0189, 0.9811], 'qtc_drug_poss_baseline'),
)
combined_qtc_drug_baseline = TensorMap(
    'combined_qtc_drug_baseline', Interpretation.CATEGORICAL,
    channel_map={'no_combined_qtc_drug_baseline': 0, 'combined_qtc_drug_baseline': 1},
    loss=weighted_crossentropy([0.0389, 0.9611], 'combined_qtc_drug_baseline'),
)

class1_baseline = TensorMap('class1_baseline', Interpretation.CATEGORICAL, channel_map={'no_class1_baseline': 0, 'class1_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0023, 0.9977], 'class1_baseline'))
bb_baseline = TensorMap('bb_baseline', Interpretation.CATEGORICAL, channel_map={'no_bb_baseline': 0, 'bb_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0453, 0.9547], 'bb_baseline'))
class3_baseline = TensorMap('class3_baseline', Interpretation.CATEGORICAL, channel_map={'no_class3_baseline': 0, 'class3_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0011, 0.9989], 'class3_baseline'))
ccb_baseline = TensorMap('ccb_baseline', Interpretation.CATEGORICAL, channel_map={'no_ccb_baseline': 0, 'ccb_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0044, 0.9956], 'ccb_baseline'))
qtc_drug_def_baseline = TensorMap('qtc_drug_def_baseline', Interpretation.CATEGORICAL, channel_map={'no_qtc_drug_def_baseline': 0, 'qtc_drug_def_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0210, 0.9790], 'qtc_drug_def_baseline'))
qtc_drug_poss_baseline = TensorMap('qtc_drug_poss_baseline', Interpretation.CATEGORICAL, channel_map={'no_qtc_drug_poss_baseline': 0, 'qtc_drug_poss_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0189, 0.9811], 'qtc_drug_poss_baseline'))
class1_fu = TensorMap('class1_fu', Interpretation.CATEGORICAL, channel_map={'no_class1_fu': 0, 'class1_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0018, 0.9982], 'class1_fu'))
bb_fu = TensorMap('bb_fu', Interpretation.CATEGORICAL, channel_map={'no_bb_fu': 0, 'bb_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0306, 0.9694], 'bb_fu'))
class3_fu = TensorMap('class3_fu', Interpretation.CATEGORICAL, channel_map={'no_class3_fu': 0, 'class3_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0006, 0.9994], 'class3_fu'))
ccb_fu = TensorMap('ccb_fu', Interpretation.CATEGORICAL, channel_map={'no_ccb_fu': 0, 'ccb_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0035, 0.9965], 'ccb_fu'))
qtc_drug_def_fu = TensorMap('qtc_drug_def_fu', Interpretation.CATEGORICAL, channel_map={'no_qtc_drug_def_fu': 0, 'qtc_drug_def_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0140, 0.9860], 'qtc_drug_def_fu'))
qtc_drug_poss_fu = TensorMap('qtc_drug_poss_fu', Interpretation.CATEGORICAL, channel_map={'no_qtc_drug_poss_fu': 0, 'qtc_drug_poss_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0127, 0.9873], 'qtc_drug_poss_fu'))
qtc_drug_def_any = TensorMap('qtc_drug_def_any', Interpretation.CATEGORICAL, channel_map={'no_qtc_drug_def_any': 0, 'qtc_drug_def_any': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0302, 0.9698], 'qtc_drug_def_any'))
qtc_drug_poss_any = TensorMap('qtc_drug_poss_any', Interpretation.CATEGORICAL, channel_map={'no_qtc_drug_poss_any': 0, 'qtc_drug_poss_any': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0267, 0.9733], 'qtc_drug_poss_any'))
any_class1 = TensorMap('any_class1', Interpretation.CATEGORICAL, channel_map={'no_any_class1': 0, 'any_class1': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0031, 0.9969], 'any_class1'))
any_bb = TensorMap('any_bb', Interpretation.CATEGORICAL, channel_map={'no_any_bb': 0, 'any_bb': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0602, 0.9398], 'any_bb'))
any_class3 = TensorMap('any_class3', Interpretation.CATEGORICAL, channel_map={'no_any_class3': 0, 'any_class3': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0013, 0.9987], 'any_class3'))
any_ccb = TensorMap('any_ccb', Interpretation.CATEGORICAL, channel_map={'no_any_ccb': 0, 'any_ccb': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0062, 0.9938], 'any_ccb'))
combined_qtc_drug_baseline = TensorMap('combined_qtc_drug_baseline', Interpretation.CATEGORICAL, channel_map={'no_combined_qtc_drug_baseline': 0, 'combined_qtc_drug_baseline': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0389, 0.9611], 'combined_qtc_drug_baseline'))
combined_qtc_drug_fu = TensorMap('combined_qtc_drug_fu', Interpretation.CATEGORICAL, channel_map={'no_combined_qtc_drug_fu': 0, 'combined_qtc_drug_fu': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0260, 0.9740], 'combined_qtc_drug_fu'))
combined_qtc_drug_any = TensorMap('combined_qtc_drug_any', Interpretation.CATEGORICAL, channel_map={'no_combined_qtc_drug_any': 0, 'combined_qtc_drug_any': 1}, loss_weight=100.0, loss=weighted_crossentropy([0.0546, 0.9454], 'combined_qtc_drug_any'))

ecg_bike_max_hr_no0 = TensorMap(
    'bike_max_hr', Interpretation.CONTINUOUS, channel_map={'bike_max_hr': 0},
    loss=ignore_zeros_logcosh, metrics=['logcosh'], normalization={'mean': 110.03, 'std': 20.04},
)
ecg_bike_resting_hr_no0 = TensorMap(
    'bike_resting_hr', Interpretation.CONTINUOUS, channel_map={'bike_resting_hr': 0},
    loss=ignore_zeros_logcosh, metrics=['logcosh'], normalization={'mean': 71.2, 'std': 12.57},
)
ecg_bike_max_pred_hr_no0 = TensorMap(
    'bike_max_pred_hr', Interpretation.CONTINUOUS, channel_map={'bike_max_pred_hr': 0},
    loss=ignore_zeros_logcosh, metrics=['logcosh'], normalization={'mean': 167.5, 'std': 5.78},
)
ecg_bike_raw_trend_hr = TensorMap(
    'trend_heartrate', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_load = TensorMap(
    'trend_load', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_grade = TensorMap(
    'trend_grade', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_artifact = TensorMap(
    'trend_artifact', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_mets = TensorMap(
    'trend_mets', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_pacecount = TensorMap(
    'trend_pacecount', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_phasename = TensorMap(
    'trend_phasename', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_phasetime = TensorMap(
    'trend_phasetime', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_time = TensorMap(
    'trend_time', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_trend_vecount = TensorMap(
    'trend_vecount', Interpretation.CONTINUOUS, shape=(87,), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_raw_full = TensorMap(
    'full', Interpretation.CONTINUOUS, shape=(216500, 3), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)
ecg_bike_median = TensorMap(
    'median', Interpretation.CONTINUOUS, shape=(5500, 3), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date, normalization=ZeroMeanStd1(), channel_map=ECG_BIKE_LEADS,
)
ecg_bike_strip = TensorMap(
    'strip', Interpretation.CONTINUOUS, shape=(5000, 3), path_prefix='ukb_ecg_bike',
    tensor_from_file=normalized_first_date,
)

def ppg_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    ppg = np.zeros(tm.shape,  dtype=np.float32)
    ppg[:, 0] = hd5[tm.name]
    return ppg

ppg_0 = TensorMap(
    'ppg_0', shape=(100, 1), tensor_from_file=ppg_from_hd5, channel_map={'ppg_0': 0},
    normalization=Standardize(mean=4824.6, std=3235.8),
)

ppg_2 = TensorMap(
    'ppg_2', shape=(100, 1), tensor_from_file=ppg_from_hd5, channel_map={'ppg_2': 0},
    normalization=Standardize(mean=4824.6, std=3235.8),
)
